/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import EDU.auburn.VGJ.util.DPoint;
import EDU.auburn.VGJ.util.DDimension;

/**
 * A ViewportScroller is a window used to scroll a window through a
 * larger area of content. The content is shown as a white rectangle, with
 * the window being represented by a black rectangle outline within it. A
 * StateChanged event is fired when the user drags the "window" with the mouse.
 * Based upon EDU.auburn.VGJ.gui.ViewportScroller. Adapted for use with Swing.
 *
 * @author Dennis Wagelaar
 * @author Larry Barowski
 * @version 0.0.16, 23 March 2002
 * @see EDU.auburn.VGJ.gui.ViewportScroller
 */

public class ViewportScroller extends JPanel
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private int width_ = -1, height_ = -1;
  private double portWidth_, portHeight_;
  private double contentWidth_, contentHeight_;
  private double offsetX_, offsetY_;
  private Rectangle portRect_;
  private Rectangle contentRect_;
  private int dragOffsetX_, dragOffsetY_;
  private Image backImage_;
  private Color backColor_ = Color.white;
  private Color lineColor_ = Color.black;
  private boolean isDragging_;
  /**
   * Only one <code>ChangeEvent</code> is needed per instance since the
   * event's only state is the source property.  The source of events
   * generated is always "this".
   */
  protected transient ChangeEvent changeEvent;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ViewPortScroller
   */
  public ViewportScroller()
  {
    super();
    try
    {
      jbInit();
      mInit();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  } //ViewPortScroller

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /* JBuilder Component initialization. */
  private void jbInit() throws Exception
  {
    this.setPreferredSize(new Dimension(100, 100));
    this.addMouseMotionListener(new java.awt.event.MouseMotionAdapter()
    {
      public void mouseDragged(MouseEvent e)
      {
        this_mouseDragged(e);
      }
    });
    this.addMouseListener(new java.awt.event.MouseAdapter()
    {
      public void mousePressed(MouseEvent e)
      {
        this_mousePressed(e);
      }
      public void mouseReleased(MouseEvent e)
      {
        this_mouseReleased(e);
      }
    });
  } //jbInit

  /* Manual initialization. */
  private void mInit()
  {
    contentRect_ = new Rectangle();
    portRect_ = new Rectangle();
    recompute();
  } //mInit

  /* Re-computes boundaries. */
  private void recompute()
  {
    double d_width = (double) width_ - 1.0;
    double d_height = (double) height_ - 1.0;

    if (d_width * contentHeight_ > d_height * contentWidth_)
    // Canvas is proportionally wider than content.
    {
      contentRect_.y = 0;
      contentRect_.height = (int) d_height;
      contentRect_.width = (int) (d_height * contentWidth_ / contentHeight_);
      contentRect_.x = (int) ((d_width - contentRect_.width) / 2.0);
    }
    else
    // Canvas is proportional with or proportionally taller thatn
    // content.
    {
      contentRect_.x = 0;
      contentRect_.width = (int) d_width;
      contentRect_.height = (int) (d_width * contentHeight_ / contentWidth_);
      contentRect_.y = (int)((d_height - contentRect_.height) / 2.0);
    }

    double scale_ratio = ((double) contentRect_.width) / contentWidth_;

    portRect_.x = contentRect_.x + (int) (offsetX_ * scale_ratio);
    portRect_.y = contentRect_.y + (int) (offsetY_ * scale_ratio);
    portRect_.width = (int) (portWidth_ * scale_ratio);
    portRect_.height = (int) (portHeight_ * scale_ratio);

    if(portRect_.x < contentRect_.x)
      portRect_.x = contentRect_.x;
    if(portRect_.y < contentRect_.y)
      portRect_.y = contentRect_.y;
    if(portRect_.x + portRect_.width > contentRect_.x + contentRect_.width)
      portRect_.x = contentRect_.x + contentRect_.width - portRect_.width;
    if(portRect_.y + portRect_.height > contentRect_.y + contentRect_.height)
      portRect_.y = contentRect_.y + contentRect_.height - portRect_.height;
  } //recompute

  /* Mouse button pressed. */
  void this_mousePressed(MouseEvent e)
  {
    isDragging_ = true;
    dragOffsetX_ = e.getX() - portRect_.x + contentRect_.x;
    dragOffsetY_ = e.getY() - portRect_.y + contentRect_.y;
  } //this_mousePressed

  /* Mouse button released. */
  void this_mouseReleased(MouseEvent e)
  {
    isDragging_ = false;

    // Post an event indicating we are done scrolling.
    fireStateChanged();
  } //this_mouseReleased

  // Mouse dragged. */
  void this_mouseDragged(MouseEvent e)
  {
    int x = e.getX();
    int y = e.getY();

    if (x < dragOffsetX_)
      x = dragOffsetX_;
    else if (x - dragOffsetX_ + portRect_.width > contentRect_.width)
      x = contentRect_.width + dragOffsetX_ - portRect_.width;
    if (y < dragOffsetY_)
      y = dragOffsetY_;
    else if (y - dragOffsetY_ + portRect_.height > contentRect_.height)
      y = contentRect_.height + dragOffsetY_ - portRect_.height;
    // save port size
    portRect_.x = x - dragOffsetX_ + contentRect_.x;
    portRect_.y = y - dragOffsetY_ + contentRect_.y;
    // paint buffer
    Graphics g = getGraphics();
    paintComponentBuffer(g);
    g.dispose();
    // save offset
    double scale_ratio = contentWidth_ / (double) (contentRect_.width);
    offsetX_ = (double) (portRect_.x - contentRect_.x) * scale_ratio;
    offsetY_ = (double) (portRect_.y - contentRect_.y) * scale_ratio;
    // Post an event indicating a scroll.
    fireStateChanged();
  } //this_mouseDragged

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Paints the viewport elements on top of the panel.
   *
   * @param g The graphics context.
   */
  protected void paintComponent(Graphics g)
  {
    // if resized, discard backbuffer
    Dimension winsize = getSize();
    if (winsize.width != width_ || winsize.height != height_)
    {
      width_ = winsize.width;
      height_ = winsize.height;
      recompute();
      backImage_ = null;
    }
    // create new backbuffer if needed
    if (backImage_ == null)
    {
      backImage_ = createImage(width_, height_);
    }
    // get backbuffer graphics context
    Graphics back = backImage_.getGraphics();
    // draw backbuffer
    try
    {
      back.setColor(getBackground());
      back.fillRect(0, 0, width_, height_);
      back.setColor(getBackColor());
      back.fillRect(contentRect_.x, contentRect_.y, contentRect_.width,
        contentRect_.height);
      back.setColor(getLineColor());
      back.setPaintMode();
      back.drawRect(contentRect_.x, contentRect_.y, contentRect_.width,
        contentRect_.height);
    }
    finally
    {
      back.dispose();
    }
    // draw screen
    paintComponentBuffer(g);
  } //paintComponent

  /**
   * Paints the viewport elements on top of the panel. Does a partial paint
   * by using the backbuffer.
   *
   * @param g The graphics context.
   */
  protected synchronized void paintComponentBuffer(Graphics g)
  {
    Graphics graphics = getComponentGraphics(g);
    try
    {
      graphics.drawImage(backImage_, 0, 0, null);
      graphics.setColor(getLineColor());
      graphics.drawRect(portRect_.x + 1, portRect_.y + 1, portRect_.width - 2,
        portRect_.height - 2);
    }
    finally
    {
      graphics.dispose();
    }
  } //paintComponentBuffer

  /**
   * Notifies all listeners that have registered interest for
   * notification on this event type.  The event instance
   * is lazily created using the parameters passed into
   * the fire method.
   *
   * @see EventListenerList
   */
  protected void fireStateChanged()
  {
    // Guaranteed to return a non-null array
    Object[] listeners = listenerList.getListenerList();
    // Process the listeners last to first, notifying
    // those that are interested in this event
    for (int i = listeners.length - 2; i >= 0; i -= 2)
    {
      if (listeners[i] == ChangeListener.class)
      {
        // Lazily create the event:
        if (changeEvent == null)
          changeEvent = new ChangeEvent(this);
        ((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);
      }
    }
  } //fireStateChanged

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the viewport size.
   *
   * @return The viewport size.
   */
  public DDimension getPortSize()
  {
    return new DDimension(portWidth_, portHeight_);
  } //getPortSize

  /**
   * Sets the viewport size.
   *
   * @param size The viewport size.
   */
  public void setPortSize(DDimension size)
  {
    if (portWidth_ == size.width && portHeight_ == size.height)
      return;

    portWidth_ = size.width;
    portHeight_ = size.height;

    recompute();
    repaint();
  } //setPortSize

  /**
   * gets the content size.
   *
   * @return The content size.
   */
  public DDimension getContentSize()
  {
    return new DDimension(contentWidth_, contentHeight_);
  } //getContentSize

  /**
   * Sets the content size.
   *
   * @param size The content size.
   */
  public void setContentSize(DDimension size)
  {
    if (contentWidth_ == size.width && contentHeight_ == size.height)
      return;

    contentWidth_ = size.width;
    contentHeight_ = size.height;

    recompute();
    repaint();
  } //setContentSize

  /**
   * Gets the viewport offset.
   *
   * return The viewport offset.
   */
  public DPoint getOffset()
  {
    return new DPoint(offsetX_, offsetY_);
  } //getOffset

  /**
   * Sets the viewport offset.
   *
   * param offset The viewport offset.
   */
  public void setOffset(DPoint offset)
  {
    if (offsetX_ == offset.x && offsetY_ == offset.y)
      return;

    offsetX_ = offset.x;
    offsetY_ = offset.y;

    recompute();
    repaint();
  } //setOffset

  /**
   * Gets the background drawing color.
   *
   * @return The background drawing color.
   */
  public Color getBackColor()
  {
    return backColor_;
  } //getBackColor

  /**
   * Sets the background drawing color.
   *
   * @param color The background drawing color.
   */
  public void setBackColor(Color color)
  {
    backColor_ = color;
  } //setBackColor

  /**
   * Gets the line drawing color.
   *
   * @return The line drawing color.
   */
  public Color getLineColor()
  {
    return lineColor_;
  } //getLineColor

  /**
   * Sets the line drawing color.
   *
   * @param color The line drawing color.
   */
  public void setLineColor(Color color)
  {
    lineColor_ = color;
  } //setLineColor

  /**
   * Returns whether the mouse is dragging.
   *
   * @return True if and only if the mouse is dragging.
   */
  public boolean isDragging()
  {
    return isDragging_;
  } //isDragging

  /**
   * Adds a ChangeListener to the viewport scroller.
   *
   * @param l the listener to be added
   */
  public void addChangeListener(ChangeListener l)
  {
    listenerList.add(ChangeListener.class, l);
  } //addChangeListener

  /**
   * Removes a ChangeListener from the viewport scroller.
   *
   * @param l the listener to be removed
   */
  public void removeChangeListener(ChangeListener l)
  {
    listenerList.remove(ChangeListener.class, l);
  } //removeChangeListener
} //class ViewPortScroller
