/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.awt.*;
import java.util.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import EDU.auburn.VGJ.graph.Node;
import EDU.auburn.VGJ.graph.GMLobject;
import EDU.auburn.VGJ.util.*;
import EDU.utwente.CoCompose.util.*;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;

/**
 * The abstract base class for CoCompose design language primitives.
 *
 * @author Dennis Wagelaar
 * @version 0.1.20, 24 March 2004
 */

public abstract class AbstractConcept extends Node
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String XMLNAME_CONCEPT     = "Concept";
  private static final String XMLNAME_GML         = "GML";
  private static final String XMLNAME_COMMENTS    = "Comments";
  private static final String XMLNAME_SPEC        = "Specification";
  /**
   * Default edge/link color.
   */
  protected static final Color DEFAULT_EDGE_COLOR   = new Color(153, 0, 51);
  /**
   * Default font color.
   */
  protected static final Color DEFAULT_FONT_COLOR   = Color.black;
  /**
   * Default selection rectangle color.
   */
  protected static final Color DEFAULT_SELECT_COLOR = Color.black;
  /**
   * Default shape filling color.
   */
  protected static final Color DEFAULT_FILL_COLOR   = new Color(255, 255, 204);
  /**
   * Default shape (oval).
   */
  protected static final int DefaultShape = OVAL;
  /**
   * Default label position.
   */
  protected static final String DefaultLabelPos = "Inside";
  /**
   * Font size (fixed).
   */
  public static final int FONT_SIZE = 12;

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private Identifier name_ = null;
  private String comments_ = new String();
  private String specification_ = new String();
  private String specificationLang_ = new String();
  private Model model_ = null;
  private Color fillColor_;
  private Color edgeColor_;
  private Color fontColor_;
  private Color selectColor_;
  private Stack copiedFrom_ = new Stack();
  // The solution role representing the role part this concept is part of.
  private SolutionRole partOf_ = null;
  // When instantiated from a solution pattern role part, this attribute says
  // what concept instantiated the role part.
  private AbstractConcept instantiatedBy_ = null;
  /**
   * The realm in which identifiers must be unique.
   */
  protected UniqueRealm realm_ = null;
  /**
   * The default label displayed in the model when no name is set.
   */
  protected String defaultLabel_ = new String();
  /**
   * Change listeners (observers).
   */
  protected Set changeListeners = new HashSet();
  
  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a Concept.
   */
  public AbstractConcept()
  {
    super();
    setDefaults();
  } //AbstractConcept

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Create XML from the GML data using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void gmlToXML(DefaultHandler handler) throws SAXException
  {
    handler.startElement(null, null, XMLNAME_GML, null);
    GMLobject gml = new GMLobject("node", GMLobject.GMLlist);
    setGMLvalues(gml);
    String gmlString = gml.toString(0);
    handler.characters(gmlString.toCharArray(), 0, gmlString.length());
    handler.endElement(null, null, XMLNAME_GML);
  } //gmlToXML

  /**
   * Create XML from the comments attribute using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void commentsToXML(DefaultHandler handler) throws SAXException
  {
    handler.startElement(null, null, XMLNAME_COMMENTS, null);
    handler.characters(getComments().toCharArray(), 0, getComments().length());
    handler.endElement(null, null, XMLNAME_COMMENTS);
  } //commentsToXML

  /**
   * Create XML from the specification attribute using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void specToXML(DefaultHandler handler) throws SAXException
  {
    AttributesImpl specAtts = new AttributesImpl();

    specAtts.addAttribute(null, null, "Language", "CDATA", getSpecificationLang());
    handler.startElement(null, null, XMLNAME_SPEC, specAtts);
    handler.characters(getSpecification().toCharArray(), 0, getSpecification().length());
    handler.endElement(null, null, XMLNAME_SPEC);
  } //specToXML

  /**
   * Create XML start from this object using a SAX DefaultHandler, element name
   * and given attributes.
   *
   * @param handler The SAX DefaultHandler.
   * @param element The XML element name.
   * @param atts The XML element attributes.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void startXML(DefaultHandler handler, String element,
    AttributesImpl atts) throws SAXException
  {
    // create new attributes list if none given.
    if (atts == null)
      atts = new AttributesImpl();

    // Write XML
    atts.addAttribute(null, null, "Id", "ID", String.valueOf(getId()));
    atts.addAttribute(null, null, "Name", "CDATA", getName());
    if (getPartOf() != null)
      atts.addAttribute(null, null, "PartOf", "IDREF",
        String.valueOf(getPartOf().getId()));
    handler.startElement(null, null, element, atts);

    // GML
    gmlToXML(handler);

    // Comments
    commentsToXML(handler);

    // Specification
    specToXML(handler);
  } //startXML

  /**
   * Create XML end from this object using a SAX DefaultHandler and element
   * name.
   *
   * @param handler The SAX DefaultHandler.
   * @param element The XML element name.
   * @param atts The XML element attributes.
   * @throws SAXException if an XML handler error occurs.
   */
  protected void endXML(DefaultHandler handler, String element)
    throws SAXException
  {
    handler.endElement(null, null, element);
  } //endXML

  /**
   * Scale the bounds of a Concept if label is put inside.
   *
   * @param graphics The graphics context.
   * @param scale The drawing scale.
   */
  protected void scaleBounds(Graphics graphics, double scale)
  {
    if (getLabelPosition() == IN)
    {
      FontMetrics fm = graphics.getFontMetrics();
      double w;
      double h;

      if (label_.length > 0)
      {
        int i;
        h = fm.getHeight() * label_.length;
        for (w = 0.0, i = 0; i < label_.length; i++)
        {
          if (fm.stringWidth(label_[i]) > w)
            w = fm.stringWidth(label_[i]);
        }
      }
      else
      {
        h = fm.getHeight();
        w = fm.stringWidth(getLabel());
      }

      if (getShape() == OVAL)
      {
        w *= 1.42;
        h *= 1.42;
      }

      w += 6;
      h += 6;

      setBoundingBox(w / scale, h / scale);
    }
  } // scaleBounds

  /**
   * Draws the shape of the concept.
   *
   * @param graphics The graphics context.
   * @param x The transformed X coordinate.
   * @param y The transformed Y coordinate.
   * @param w The width.
   * @param h The height.
   */
  protected void drawShape(Graphics graphics, int x, int y, double w, double h)
  {
    if (getShape() == OVAL)
    {
      graphics.setColor(getFillColor());
      graphics.fillOval((int)(x - w / 2), (int)(y - h / 2), (int)w, (int)h);
      graphics.setColor(getEdgeColor());
      graphics.drawOval((int)(x - w / 2), (int)(y - h / 2), (int)w, (int)h);
      if(isGroup_ && w > 4 && h > 4)
        graphics.drawOval((int)(x - w / 2 + 2), (int)(y - h / 2 + 2),
          (int)w - 4, (int)h - 4);
    }
    else if (getShape() == RECTANGLE)
    {
      graphics.setColor(getFillColor());
      graphics.fillRect((int)(x - w / 2), (int)(y - h / 2), (int)w, (int)h);
      graphics.setColor(getEdgeColor());
      graphics.drawRect((int)(x - w / 2), (int)(y - h / 2), (int)w, (int)h);
      if (isGroup_ && w > 4 && h > 4)
        graphics.drawRect((int)(x - w / 2 + 2), (int)(y - h / 2 + 2),
          (int)w - 4, (int)h - 4);
    }
  } //drawShape

  /**
   * Draws the label of the concept.
   *
   * @param graphics The graphics context.
   * @param x The transformed X coordinate.
   * @param y The transformed Y coordinate.
   * @param h The height of the concept shape.
   */
  protected void drawLabel(Graphics graphics, int x, int y, double h)
  {
    FontMetrics fm;
    double width;
    double lx, ly;

    fm = graphics.getFontMetrics();
    graphics.setColor(getFontColor());
    // determine position
    if (getLabelPosition() != BELOW)
    {
      if (label_.length > 0)
        ly = y - fm.getHeight() * (label_.length - 2) / 2.0;
      else
        ly = y + fm.getHeight() / 2.0;
      ly -= fm.getDescent();
    }
    else
    {
      ly = y + h / 2.0 + fm.getAscent() + 1.0;
    }
    // draw label
    if (label_.length == 0)
    {
      width = fm.stringWidth(getLabel());
      lx = x - width / 2.0;
      graphics.drawString(getLabel(), (int) lx, (int) ly);
    }
    else
    {
      for (int i = 0; i < label_.length; i++)
      {
        width = fm.stringWidth(label_[i]);
        lx = x - width / 2.0;
        graphics.drawString(label_[i], (int) lx, (int) ly);
        ly += fm.getHeight();
      }
    }
  } // drawLabel

  /**
   * Draws the selection artifacts of the concept.
   *
   * @param graphics The graphics context.
   * @param position The transformed coordinates.
   * @param scale The drawing scale.
   */
  protected void drawSelected(Graphics graphics, DPoint3 position, double scale)
  {
    DDimension bbox = getBoundingBox();
    DPoint tr = new DPoint(position.x + bbox.width * scale / 2.0 + 1,
      position.y - bbox.height * scale / 2.0 - 1);
    DPoint bl = new DPoint(position.x - bbox.width * scale / 2.0 - 1,
      position.y + bbox.height * scale / 2.0 + 1);

    // Draw selection components.
    graphics.setColor(getSelectColor());

    graphics.drawRect((int)position.x - 4, (int)position.y - 1, 8, 2);
    graphics.drawRect((int)position.x - 1, (int)position.y - 4, 2, 8);

    graphics.drawRect((int)tr.x - 2, (int)tr.y - 2, 4, 4);

    graphics.drawRect((int)position.x - 3, (int)bl.y - 1, 6, 2);
    graphics.drawRect((int)bl.x - 1, (int)position.y - 3, 2, 6);

    graphics.setColor(Color.white);
    graphics.drawLine((int)position.x - 3, (int)position.y,
      (int)position.x + 3, (int)position.y);
    graphics.drawLine((int)position.x, (int)position.y - 3,
      (int)position.x, (int)position.y + 3);
    graphics.drawRect((int)tr.x - 1, (int)tr.y - 1, 2, 2);
    graphics.drawLine((int)position.x - 2, (int)bl.y,
      (int)position.x + 2, (int)bl.y);
    graphics.drawLine((int)bl.x, (int)position.y - 2,
      (int)bl.x, (int)position.y + 2);
  } //drawSelected

  /**
   * Aligns new coordinates to the grid, if needed.
   *
   * @param new_position The new coordinates.
   */
  protected void alignToGrid(DPoint new_position)
  {
    // if needs alignment to grid, align
    if (getModel().getAlignToGrid())
    {
      new_position.x = Math.round(new_position.x / 10) * 10;
      new_position.y = Math.round(new_position.y / 10) * 10;
    }
  } //alignToGrid

  /**
   * Aligns new coordinates to the grid, if needed.
   *
   * @param new_position The new coordinates.
   */
  protected void alignToGrid(DPoint3 new_position)
  {
    // if needs alignment to grid, align
    if (getModel().getAlignToGrid())
    {
      new_position.x = Math.round(new_position.x / 10) * 10;
      new_position.y = Math.round(new_position.y / 10) * 10;
      new_position.z = Math.round(new_position.z / 10) * 10;
    }
  } //alignToGrid

  /**
   * Notifies the PropertyChangeListeners of a PropertyChangeEvent.
   * @param evt The PropertyChangeEvent to send.
   */
  protected void notifyListeners(PropertyChangeEvent evt)
  {
    HashSet listenersCopy = new HashSet(changeListeners); 
    Iterator listeners = listenersCopy.iterator();
    while (listeners.hasNext())
    {
      ((PropertyChangeListener) listeners.next()).propertyChange(evt);
    }
  } //notifyListeners

  /**
   * Notfies the PropertyChangeListeners of a name change.
   * @param oldName The old name property value.
   * @param newName The new name property value.
   */
  protected void notifyNameChange(String oldName, String newName)
  {
    notifyListeners(new PropertyChangeEvent(this, "Name", oldName, newName));
  } //notifyNameChange

  /**
   * Notfies the PropertyChangeListeners that we are removed from this model.
   */
  protected void notifyRemovedFromModel()
  {
    notifyListeners(new PropertyChangeEvent(this, "Removed", null, null));
  } //notifyNameChange

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Checks whether this concept has been copied from the given AbstractConcept.
   * @param from the AbstractConcept to check for.
   * @return true if this AbstractConcept has been copied from the given AbstractConcept.
   */
  public boolean isCopiedFrom(AbstractConcept from)
  {
    return copiedFrom_.contains(from);
  }
  
  /**
   * Returns the AbstractConcept from which this is last copied, if any.
   * @return the AbstractConcept from which this is last copied, if any.
   */
  public AbstractConcept getLastCopiedFrom()
  {
    if (copiedFrom_.size() > 0)
      return (AbstractConcept) copiedFrom_.peek();
    else
      return null;
  }

  /**
   * Returns the stack of AbstractConcepts from which this has been copied.
   * @return the stack of AbstractConcepts from which this has been copied.
   */  
  public Stack getCopiedFrom()
  {
    return copiedFrom_;
  }

  /**
   * Sets the defaults for this concept.
   */
  public void setDefaults()
  {
    defaultLabel_ = "Concept";
    setShape(DefaultShape);
    setLabelPosition(DefaultLabelPos);
    // set default colors
    setEdgeColor(DEFAULT_EDGE_COLOR);
    setFillColor(DEFAULT_FILL_COLOR);
    setFontColor(DEFAULT_FONT_COLOR);
    setSelectColor(DEFAULT_SELECT_COLOR);
  } //setDefaults

  /**
   * Sets a new position in the model.
   *
   * @param new_x The new X coordinate.
   * @param new_y The new Y coordinate.
   */
  public void setPosition(double new_x, double new_y)
  {
    DPoint new_position = new DPoint(new_x, new_y);
    setPosition(new_position);
  } //setPosition(double new_x, double new_y)

  /**
   * Sets a new position in the model.
   *
   * @param new_position The new coordinates.
   */
  public synchronized void setPosition(DPoint new_position)
  {
    // align if needed
    alignToGrid(new_position);
    // inherited setPosition
    super.setPosition(new_position);
  } //setPosition(DPoint new_position)

  /**
   * Sets a new position in the model.
   *
   * @param new_x The new X coordinate.
   * @param new_y The new Y coordinate.
   * @param new_z The new Z coordinate.
   */
  public void setPosition(double new_x, double new_y, double new_z)
  {
    DPoint new_position = new DPoint(new_x, new_y);
    setPosition(new_position);
  } //setPosition(double new_x, double new_y, double new_z)

  /**
   * Sets a new position in the model.
   *
   * @param new_position The new coordinates.
   */
  public synchronized void setPosition(DPoint3 new_position)
  {
    // align if needed
    alignToGrid(new_position);
    // inherited setPosition
    super.setPosition(new_position);
  } //setPosition(DPoint3 new_position)

  /**
   * Moves to a new position in the model.
   *
   * @param moveTransform Move transformation matrix - see Node class.
   * @param viewTransform View transformation matrix - see Node class.
   * @param xoffs X offset - see Node class.
   * @param yoffs Y offset - see Node class.
   */
  public void slide(Matrix44 moveTransform, Matrix44 viewTransform,
    int xoffs, int yoffs)
  {
    super.slide(moveTransform, viewTransform, xoffs, yoffs);
    setPosition(getPosition3());
  } //slide

  /**
   * Sets the uniqueness realm.
   *
   * @param realm The uniqueness realm for identifiers.
   */
  public void setUniqueRealm(UniqueRealm realm)
  {
    realm_ = realm;
  } //setUniqueRealm

  /**
   * Gets the uniqueness realm.
   *
   * @return The uniqueness realm for identifiers.
   */
  public UniqueRealm getUniqueRealm()
  {
    return realm_;
  } //getUniqueRealm

  /**
   * Sets the label using the object attributes.
   */
  public synchronized void setLabel()
  {
    // set label depending on name.
    if (getName().length() > 0)
    {
      setLabel(getName());
    }
    else
    {
      setLabel(defaultLabel_ + " " + getId());
    }
  } //setLabel

  /**
   * Sets the model this concept is part of.
   *
   * @param model The model this concept is part of.
   */
  public void setModel(Model model)
  {
    if (model == null)
      notifyRemovedFromModel();
    model_ = model;
  } //setModel

  /**
   * Gets the model this concept is part of.
   *
   * @return The model this concept is part of.
   */
  public Model getModel()
  {
    return model_;
  } //getModel

  /**
   * Gets default fill color, depending on whether this concept is already
   * part of a role concept or not.
   *
   * @return The default fill color.
   */
  public Color getDefaultFillColor()
  {
    if (getPartOf() != null)
      return getPartOf().getDefaultFillColor();
    else
      return DEFAULT_FILL_COLOR;
  } //getDefaultFillColor

  /**
   * Sets the filling color for drawing.
   *
   * @param fillColor The color.
   */
  public void setFillColor(Color fillColor)
  {
    fillColor_ = fillColor;
  } //setFillColor

  /**
   * Gets the filling color for drawing.
   *
   * @return The filling color for drawing.
   */
  public Color getFillColor()
  {
    return fillColor_;
  } //getFillColor

  /**
   * Sets the edge color for drawing.
   *
   * @param edgeColor The color.
   */
  public void setEdgeColor(Color edgeColor)
  {
    edgeColor_ = edgeColor;
  } //setEdgeColor

  /**
   * Gets the edge color for drawing.
   *
   * @return The edge color for drawing.
   */
  public Color getEdgeColor()
  {
    return edgeColor_;
  } //getEdgeColor

  /**
   * Sets the font color for drawing.
   *
   * @param fontColor The color.
   */
  public void setFontColor(Color fontColor)
  {
    fontColor_ = fontColor;
  } //setFontColor

  /**
   * Gets the font color for drawing.
   *
   * @return The font color for drawing.
   */
  public Color getFontColor()
  {
    return fontColor_;
  } //getFontColor

  /**
   * Sets the selection artifact color for drawing.
   *
   * @param selectColor The color.
   */
  public void setSelectColor(Color selectColor)
  {
    selectColor_ = selectColor;
  } //setSelectColor

  /**
   * Gets the selection artifact color for drawing.
   *
   * @return The select color for drawing.
   */
  public Color getSelectColor()
  {
    return selectColor_;
  } //getSelectColor

  /**
   * Sets Id and initial label as well.
   *
   * @param id Node identifier.
   */
  public void setId(int id)
  {
    super.setId(id);
    setLabel();
  } //setId

  /**
   * Gets name attribute.
   *
   * @return The name attribute.
   */
  public String getName()
  {
    if (name_ != null)
      return name_.getValue();
    else
      return "";
  } //getName

  /**
   * Gets the public name attribute, which is the same as the name attribute.
   *
   * @return The public name attribute.
   */
  public String getPublicName()
  {
    return getName();
  } //getPublicName

  /**
   * Sets name attribute - sets label as well.
   *
   * @param name_in The new name.
   * @throws CoComposeException if the name is not a valid unique identifier.
   */
  public synchronized void setName(String name_in) throws CoComposeException
  {
    if (name_ == null)
    {
      notifyNameChange("", name_in);
      if (name_in.length() > 0)
        name_ = new Identifier(realm_, name_in);
    }
    else
    {
      notifyNameChange(name_.getValue(), name_in);
      if (name_in.length() == 0)
      {
        if (realm_ != null)
          realm_.unRegister(name_.getValue());
        name_ = null;
      }
      else
      {
        name_.setValue(name_in);
      }
    }
    setLabel();
  } //setName

  /**
   * Gets comments attribute.
   *
   * @return The comments attribute.
   */
  public String getComments()
  {
    return comments_;
  } //getComments

  /**
   * Sets comments attribute.
   *
   * @param comments_in The new comments.
   */
  public void setComments(String comments_in)
  {
    comments_ = comments_in;
  } //setComments

  /**
   * Gets specification attribute.
   *
   * @return The specification attribute.
   */
  public String getSpecification()
  {
    return specification_;
  } //getSpecification

  /**
   * Sets specification attribute.
   *
   * @param specification_in The new specification.
   */
  public void setSpecification(String specification_in)
  {
    specification_ = specification_in;
  } //setSpecification

  /**
   * Gets specification language attribute.
   *
   * @return The specification language attribute.
   */
  public String getSpecificationLang()
  {
    return specificationLang_;
  } //getSpecificationLang

  /**
   * Sets specification language attribute.
   *
   * @param specificationLang_in The new specification language.
   */
  public void setSpecificationLang(String specificationLang_in)
  {
    specificationLang_ = specificationLang_in;
  } //setSpecificationLang

  /**
   * Sets the solution role representing the role part this concept is part of.
   *
   * @param role The solution role this concept is part of.
   */
  public synchronized void setPartOf(SolutionRole role)
  {
    // if partOf was previously set, remove this concept from its list.
    if (partOf_ != null)
    {
      partOf_.removeConcept(this);
    }
    partOf_ = role;
    // if partOf is set, add this concept to its list.
    if (role != null)
    {
      role.addConcept(this);
    }
    // set new color
    setFillColor(getDefaultFillColor());
  } //setPartOf

  /**
   * Gets the solution role representing the role part this concept is part of.
   *
   * @return The solution role this concept is part of.
   */
  public SolutionRole getPartOf()
  {
    return partOf_;
  } //getPartOf

  /**
   * Copies attributes from fromConcept to self.
   *
   * @param fromConcept The Concept to copy the attributes from.
   */
  public void copyAttributes(AbstractConcept fromConcept)
  {
    super.copyAttributes(fromConcept);
    // don't copy certain attributes, but use the default
    setDefaults();
    try
    {
      setName(fromConcept.getName());
    }
    catch (Exception e)
    { // should never happen
      e.printStackTrace();
    }
    setComments(fromConcept.getComments());
    setSpecification(fromConcept.getSpecification());
    setSpecificationLang(fromConcept.getSpecificationLang());
    copiedFrom_.addAll(fromConcept.copiedFrom_);
    copiedFrom_.push(fromConcept);
  } //copyAttributes

  /**
   * Draws the Concept on a component.
   *
   * @param comp The component to draw upon.
   * @param graphics The graphics context.
   * @param transform The coordinate transformation matrix.
   * @param quality The drawing quality.
   */
  public synchronized void draw(Component comp, Graphics graphics,
    Matrix44 transform, int quality)
  {
    // if in a group, don't draw - inherited.
    if ((isGroup_ && !groupActive_) || inActiveGroup_)
      return;

    // get scale
    double scale = transform.scale;

    // We scale the bounds here.
    Font font = new Font("Dialog", Font.PLAIN, (int) Math.round(FONT_SIZE * scale));
    graphics.setFont(font);
    scaleBounds(graphics, scale);

    double w = getBoundingBox().width * scale;
    double h = getBoundingBox().height * scale;

    if (w < 1) w = 1;
    if (h < 1) h = 1;

    DPoint3 position = getPosition3();
    position.transform(transform);

    int x = (int) position.x;
    int y = (int) position.y;

    // draw the shape
    drawShape(graphics, x, y, w, h);

    // draw label if high quality requested
    if (quality > 0)
      drawLabel(graphics, x, y, h);

    // if not selected, we're done drawing
    if (!getSelected())
      return;

    // Draw selection components.
    drawSelected(graphics, position, scale);
  } //draw

  /**
   * Create XML from this object using a SAX DefaultHandler.
   *
   * @param handler The SAX DefaultHandler.
   * @throws SAXException if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler) throws SAXException
  {
    // Write XML
    startXML(handler, XMLNAME_CONCEPT, null);
    endXML(handler, XMLNAME_CONCEPT);
  } //toXML

  /**
   * Unregisters all identifiers from the realm.
   */
  public void unRegister()
  {
    if (realm_ != null)
      realm_.unRegister(getName());
  } //unRegister

  /**
   * Generate Postscript code for the node.
   *
   * @param transform The coordinate transformation matrix (3D -> 2D).
   * @return The Postscript representation.
   */
  public String toPS(Matrix44 transform)
  {
    double scale = transform.scale;
    String result = new String();
    DDimension size = getBoundingBox();
    double w = size.width * scale;
    double h = size.height * scale;

    // clip
    if (w < 1)
      w = 1;
    if (h < 1)
      h = 1;
    // get 2D position
    DPoint3 pos = getPosition3();
    pos.transform(transform);
    // output shape
    int shape = getShape();
    String shapeStr = "ellipse\n"; // default shape
    if (shape == RECTANGLE)
    {
      shapeStr = "rectangle\n";
    }
    Color edgeColor = getEdgeColor();
    Color fillColor = getFillColor();
    double edgeR = edgeColor.getRed() / 255.0;
    double edgeG = edgeColor.getGreen() / 255.0;
    double edgeB = edgeColor.getBlue() / 255.0;
    double fillR = fillColor.getRed() / 255.0;
    double fillG = fillColor.getGreen() / 255.0;
    double fillB = fillColor.getBlue() / 255.0;
    result += PSnum_(edgeR) + PSnum_(edgeG) + PSnum_(edgeB) +
              PSnum_(fillR) + PSnum_(fillG) + PSnum_(fillB) +
              PSnum_(pos.x) + PSnum_(pos.y) + PSnum_(w) + PSnum_(h) + shapeStr;
    // output label text
    result += "[\n";
    if (label_.length > 0)
    {
      for (int i = 0; i < label_.length; i++)
      {
        result += "  (" + psString_(label_[i]) + ")\n";
      }
    }
    else
    {
      result += "  (" + psString_(getLabel()) + ")\n";
    }
    result += "]\n";
    // position label
    int labelPosition = getLabelPosition();
    if (labelPosition == BELOW)
      result += PSnum_(pos.x) + PSnum_(pos.y + h/2.0) + "0 ";
    else if (labelPosition == CENTER)
      result += PSnum_(pos.x) + PSnum_(pos.y) + "1 ";
    else
    {
      if (shape == OVAL)
      {
        w /= 1.42;
        h /= 1.42;
      }
      result += PSnum_(pos.x) + PSnum_(pos.y) + PSnum_(w) + PSnum_(h) + "2 ";
    }
    // set font color
    Color fontColor = getFontColor();
    double fontR = fontColor.getRed() / 255.0;
    double fontG = fontColor.getGreen() / 255.0;
    double fontB = fontColor.getBlue() / 255.0;
    result += PSnum_(fontR) + PSnum_(fontG) + PSnum_(fontB) + " label\n\n";
    return result;
  } //toPS

  /**
   * Get the drawing bounds (including label).
   *
   * @param graphics The graphics context to measure for.
   * @param transform The coordinate transformation matrix.
   * @param width The variable to store the width.
   * @param height The variable to store the height.
   */
  public void getDrawBounds(Graphics graphics, Matrix44 transform,
    DPoint width, DPoint height)
  {
    double labelW = 0;
    double labelH = 0;
    double cwidth, w, h;
    double scale = transform.scale;

    // We scale the bounds here.
    Font font = new Font("Dialog", Font.PLAIN, (int) Math.round(FONT_SIZE * scale));
    graphics.setFont(font);
    scaleBounds(graphics, scale);

    // get bounding box
    DDimension size = getBoundingBox();
    // determine size
    if (getLabelPosition() == IN)
    {
      w = size.width * scale;
      h = size.height * scale;
    }
    else
    {
      // get label width
      FontMetrics fm = graphics.getFontMetrics();
      if (label_.length > 0)
      {
        for (int i = 0; i < label_.length; i++)
        {
          cwidth = fm.stringWidth(label_[i]);
          if (cwidth > labelW)
            labelW = cwidth;
          labelH += fm.getHeight();
        }
      }
      else
      {
        labelW = fm.stringWidth(getLabel());
        labelH = fm.getHeight();
      }
      w = Math.max(size.width * scale, labelW * scale);
      h = (size.height + labelH) * scale;
    }
    // get position
    DPoint3 pos = getPosition3();
    pos.transform(transform);
    // calculate drawing bounds
    width.x = pos.x - w / 2.0;
    width.y = pos.x + w / 2.0;
    height.x = pos.y - h / 2.0;
    height.y = pos.y + h / 2.0;
  } //getDrawBounds

  /**
   * Gets target language to translate the concept to. Retrieved from model.
   *
   * @return The target language.
   */
  public String getTargetLanguage()
  {
    Model model = getModel();
    if (model != null)
    {
      return model.getTargetLanguage();
    }
    else
    {
      return "";
    }
  } //getTargetLanguage

  /**
   * Adds a PropertyChangeListener (observer) to the list of listeners.
   *
   * @param l The PropertyChangeListener to add.
   */
  public void addPropertyChangeListener(PropertyChangeListener l)
  {
    changeListeners.add(l);
  } //addPropertyChangeListener

  /**
   * Removes a PropertyChangeListener (observer) from the list of listeners.
   *
   * @param l The PropertyChangeListener to remove.
   */
  public void removePropertyChangeListener(PropertyChangeListener l)
  {
    changeListeners.remove(l);
  } //addPropertyChangeListener

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    visitor.visitAbstractConcept(this);
  } //accept

  /**
   * When instantiated from a solution pattern role part, returns what
   * concept instantiated the role part. Otherwise returns null.
   * Note that this is a volatile property and won't be saved with a model.
   * @return The concept that instantiated this or null.
   */
  public AbstractConcept getInstantiatedBy()
  {
    return instantiatedBy_;
  } //getInstantiatedBy

  /**
   * Sets what concept instantiated the role part to which this belongs.
   * Note that this is a volatile property and won't be saved with a model.
   * @param instantiatedBy The concept that instantiated this.
   */
  public void setInstantiatedBy(AbstractConcept instantiatedBy)
  {
    this.instantiatedBy_ = instantiatedBy;
  } //setInstantiatedBy
  
  /**
   * @see java.lang.Object#toString()
   */
  public String toString()
  {
    return getClass().getName() + "(" + getName() + ")";
  } //toString
} //class AbstractConcept
