/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import java.awt.Color;
import EDU.auburn.VGJ.util.*;
import EDU.utwente.CoCompose.util.CoComposeException;

/**
 * The abstract base class for concepts that are part of a composite.
 *
 * @author Dennis Wagelaar
 * @version 0.1.10, 22 July 2003
 */

public abstract class CompositePartConcept extends ConstrainedConcept
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Default shape filling color.
   */
  protected static final Color DEFAULT_FILL_COLOR   = Color.white;

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // The composite this concept is part of.
  private Composite composite_;
  // The offset coordinates relative to the parent composite.
  private DPoint3 compositeOffset_;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a CompositePartConcept.
   */
  public CompositePartConcept()
  {
    super();
  } //CompositePartConcept

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Bounds one dimension of a position only.
   *
   * @param new_pos A new position coordinate dimension.
   * @param size The corresponding dimension size.
   * @param bound_pos The boundary position coordinate dimension.
   * @param bound_size The corresponding dimension boundary size.
   * @return The clipped position coordinate dimension.
   */
  protected double boundDimension(double new_pos, double size,
    double bound_pos, double bound_size)
  {
    if (new_pos - size / 2 < bound_pos - bound_size / 2)
    {
      new_pos = bound_pos - bound_size / 2 + size / 2;
    }
    else if (new_pos + size / 2 > bound_pos + bound_size / 2)
    {
      new_pos = bound_pos + bound_size / 2 - size / 2;
    }
    return new_pos;
  } //boundDimension

  /**
   * Bounds a new role position within the bounds of the composite.
   *
   * @param new_position The new position coordinate to clip.
   * @return The clipped position coordinate.
   */
  protected DPoint boundPosition(DPoint new_position)
  {
    Composite composite = getComposite();

    if (composite != null)
    {
      // get composite position and size
      DDimension dim = getBoundingBox();
      DPoint compositePos = composite.getPosition();
      DDimension compositeDim = composite.getBoundingBox();
      // check x coordinate bounds
      new_position.x = boundDimension(new_position.x, dim.width, compositePos.x,
        compositeDim.width);
      // check y coordinate bounds
      new_position.y = boundDimension(new_position.y, dim.height, compositePos.y,
        compositeDim.height);
    }
    return new_position;
  } //boundPosition(DPoint new_position)

  /**
   * Bounds a new role position within the bounds of the composite.
   *
   * @param new_position The new position coordinate to clip.
   * @return The clipped position coordinate.
   */
  protected DPoint3 boundPosition(DPoint3 new_position)
  {
    Composite composite = getComposite();

    if (composite != null)
    {
      // get composite position and size
      DDimension3 dim = getBoundingBox3();
      DPoint3 compositePos = composite.getPosition3();
      DDimension3 compositeDim = composite.getBoundingBox3();
      // check x coordinate bounds
      new_position.x = boundDimension(new_position.x, dim.width, compositePos.x,
        compositeDim.width);
      // check y coordinate bounds
      new_position.y = boundDimension(new_position.y, dim.height, compositePos.y,
        compositeDim.height);
      // check z coordinate bounds
      new_position.z = boundDimension(new_position.z, dim.depth, compositePos.z,
        compositeDim.depth);
    }
    return new_position;
  } //boundPosition(DPoint3 new_position)

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets default fill color.
   *
   * @return The default fill color.
   */
  public Color getDefaultFillColor()
  {
    return DEFAULT_FILL_COLOR;
  } //getDefaultFillColor

  /**
   * Sets the Composite this PublishedConcept is part of.
   *
   * @param composite_in The Composite to link to.
   */
  public void setComposite(Composite composite_in)
  {
    composite_ = composite_in;
    setUniqueRealm(composite_in);
  } //setComposite

  /**
   * Gets the Composite this PublishedConcept is part of.
   *
   * @return The Composite this role is a part of.
   */
  public Composite getComposite()
  {
    return composite_;
  } //getComposite

  /**
   * Saves the relative offset coordinates to the parent composite, if any.
   */
  public void saveCompositeOffset()
  {
    Composite composite = getComposite();
    if (composite != null)
    {
      DPoint3 pos = composite.getPosition3();
      compositeOffset_ = getPosition3();
      compositeOffset_.x -= pos.x;
      compositeOffset_.y -= pos.y;
      compositeOffset_.z -= pos.z;
    }
  } //saveCompositeOffset

  /**
   * Restore the offset coordinates to the composite, if any.
   */
  public void restoreCompositeOffset()
  {
    Composite composite = getComposite();
    // set label as well
    setLabel();
    if ((composite != null) && (compositeOffset_ != null))
    {
      DPoint3 fpos = composite.getPosition3();
      setPosition(fpos.x + compositeOffset_.x,
                  fpos.y + compositeOffset_.y,
                  fpos.z + compositeOffset_.z);
    }
  } //restoreCompositeOffset

  /**
   * Sets the solution role the Composite is part of.
   *
   * @param role The solution role the Composite is part of.
   */
  public void setPartOf(SolutionRole role)
  {
    Composite composite = getComposite();
    if (composite != null)
      composite.setPartOf(role);
  } //setPartOf

  /**
   * Gets the solution role the Composite is part of.
   *
   * @return The solution role the Composite is part of.
   */
  public SolutionRole getPartOf()
  {
    Composite composite = getComposite();
    if (composite != null)
      return composite.getPartOf();
    else
      return null;
  } //getPartOf

  /**
   * Clips and sets a new position in the model.
   *
   * @param new_x The new X coordinate.
   * @param new_y The new Y coordinate.
   */
  public void setPosition(double new_x, double new_y)
  {
    DPoint new_position = new DPoint(new_x, new_y);
    new_position = boundPosition(new_position);
    super.setPosition(new_position);
  } //setPosition(double new_x, double new_y)

  /**
   * Clips and sets a new position in the model.
   *
   * @param new_position The new coordinates.
   */
  public void setPosition(DPoint new_position)
  {
    new_position = boundPosition(new_position);
    super.setPosition(new_position);
  } //setPosition(DPoint new_position)

  /**
   * Clips and sets a new position in the model.
   *
   * @param new_x The new X coordinate.
   * @param new_y The new Y coordinate.
   * @param new_z The new Z coordinate.
   */
  public void setPosition(double new_x, double new_y, double new_z)
  {
    DPoint3 new_position = new DPoint3(new_x, new_y, new_z);
    new_position = boundPosition(new_position);
    super.setPosition(new_position);
  } //setPosition(double new_x, double new_y, double new_z)

  /**
   * Clips and sets a new position in the model.
   *
   * @param new_position The new coordinates.
   */
  public void setPosition(DPoint3 new_position)
  {
    new_position = boundPosition(new_position);
    super.setPosition(new_position);
  } //setPosition(DPoint3 new_position)

  /**
   * Accepts a ModelVisitor.
   *
   * @param visitor The visitor object to accept.
   * @throws CoComposeException if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    super.accept(visitor);
    visitor.visitCompositePartConcept(this);
  } //accept

  /**
   * When instantiated from a solution pattern role part, returns what
   * concept instantiated the role part. Otherwise returns null.
   * This property is linked to the corresponding composite property.
   * @return The concept that instantiated this or null.
   */
  public AbstractConcept getInstantiatedBy()
  {
    return getComposite().getInstantiatedBy();
  } //getInstantiatedBy

  /**
   * Sets what concept instantiated the role part to which this belongs.
   * This property is linked to the corresponding composite property.
   * @param instantiatedBy The concept that instantiated this.
   */
  public void setInstantiatedBy(AbstractConcept instantiatedBy)
  {
    getComposite().setInstantiatedBy(instantiatedBy);
  } //setInstantiatedBy
} //class CompositePartConcept
