/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.lang;

import EDU.utwente.CoCompose.gui.ModelUpdate;
import EDU.utwente.CoCompose.translate.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Abstract Visitor class for visiting CoCompose Models.
 *
 * @author Dennis Wagelaar
 * @version 0.1.15, 20 October 2003
 */

public abstract class ModelVisitor implements Runnable
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // Thread object to run visitor in thread
  private Thread thread_;
  // Model cache for starting the visitor through a thread
  private Model model_;
  // Model display for controlling whether or not user can work on the model
  // during visiting operation.
  private ModelUpdate update_;
  // Error reporter
  private ErrorReporter report_;
  /**
   * The object to log to.
   */
  protected Logger log = null;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ModelVisitor.
   */
  public ModelVisitor()
  {
  } //ModelVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the updateable model display.
   *
   * @return The updateable model display.
   */
  protected ModelUpdate getUpdate()
  {
    return update_;
  } //getUpdate

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the logger object.
   *
   * @param logger The object to log to
   */
  public void setLogger(Logger logger)
  {
    log = logger;
  } //setLogger

  /**
   * Sets the error reporter object. This is necessary when the visitor is
   * executed in a separate thread.
   *
   * @param reporter The error reporter.
   */
  public void setErrorReporter(ErrorReporter reporter)
  {
    report_ = reporter;
  } //setErrorReporter

  /**
   * Sets the updateable model display.
   *
   * @param update The updateable model display.
   */
  public void setUpdate(ModelUpdate update)
  {
    update_ = update;
  } //setUpdate

  /**
   * Starts visitor in a separate thread.
   *
   * @param model The model to visit.
   */
  public void startModel(Model model)
  {
    model_ = model;
    thread_ = new Thread(this);
    thread_.setPriority(Thread.MIN_PRIORITY);
    thread_.start();
  } //start

  /**
   * Runs the visitor. This method should not be called directly. Call start
   * instead. Note that exceptions cannot be thrown, so they are printed.
   */
  public void run()
  {
    try
    { // allow starting thread some recovery time before starting this one
      Thread.sleep(1000);
    }
    catch (InterruptedException ie)
    { // should never happen
      ie.printStackTrace();
      return;
    }

    ModelUpdate update = getUpdate();

    if (model_ == null)
      return;
    try
    { // disable display
      if (update != null)
        update.setEnabled(false);
      try
      { // start visitor
        model_.accept(this);
        // notify we're done
        if (report_ != null)
          report_.notify(this);
      }
      finally
      { // enable display
        if (update != null)
          update.setEnabled(true);
      }
    }
    catch (CoComposeException e)
    { // notify we're done if no error exception
      if ((report_ != null) && (!e.getType().equals("Error")))
        report_.notify(this);
      if (report_ != null)
        report_.reportError(e);
      else
        e.printStackTrace();
    }
  } //run

  /**
   * Visits a Model object.
   *
   * @param model The model to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitModel(Model model) throws CoComposeException
  {
  } //visitModel

  /**
   * Visits an AbstractConcept object.
   *
   * @param concept The abstract concept to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitAbstractConcept(AbstractConcept concept) throws CoComposeException
  {
  } //visitAbstractConcept

  /**
   * Visits a ConstrainedConcept object.
   *
   * @param concept The constrained concept to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitConstrainedConcept(ConstrainedConcept concept)
    throws CoComposeException
  {
  } //visitConstrainedConcept

  /**
   * Visits a CompositePartConcept object.
   *
   * @param concept The composite part concept to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitCompositePartConcept(CompositePartConcept concept)
    throws CoComposeException
  {
  } //visitCompositePartConcept

  /**
   * Visits a Concept object.
   *
   * @param concept The concept to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitConcept(Concept concept) throws CoComposeException
  {
  } //visitConcept

  /**
   * Visits a SolutionRole object.
   *
   * @param role The role to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitSolutionRole(SolutionRole role) throws CoComposeException
  {
  } //visitSolutionRole

  /**
   * Visits a Composite object.
   *
   * @param composite The composite to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
  } //visitComposite

  /**
   * Visits a CompositeRole object.
   *
   * @param role The composite role to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitCompositeRole(CompositeRole role) throws CoComposeException
  {
  } //visitCompositeRole

  /**
   * Visits a PublishedConcept object.
   *
   * @param role The composite role to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitPublishedConcept(PublishedConcept role) throws CoComposeException
  {
  } //visitPublishedConcept

  /**
   * Visits a Link object.
   *
   * @param link The link to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitLink(Link link) throws CoComposeException
  {
  } //visitLink

  /**
   * Visits an Implementation object.
   *
   * @param implementation The concept implementation to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitImplementation(Implementation implementation)
    throws CoComposeException
  {
  } //visitImplementation

  /**
   * Visits an ImplementationGenerator object.
   *
   * @param implementation The implementation generator to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitImplementationGenerator(ImplementationGenerator implementation)
    throws CoComposeException
  {
  } //visitImplementationGenerator

  /**
   * Visits an ImplementationRolePart object.
   *
   * @param rolepart The implementation generator role part to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitImplementationRolePart(ImplementationRolePart rolepart)
    throws CoComposeException
  {
  } //visitImplementationRolePart

  /**
   * Visits a SolutionPattern object.
   *
   * @param solution The solution pattern to visit.
   * @throws CoComposeException if visiting operation failed.
   */
  public void visitSolutionPattern(SolutionPattern solution) throws CoComposeException
  {
  } //visitSolutionPattern
} //class ModelVisitor
