/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Checks the concept form constraints of the model. Supports incremental
 * checking, i.e. only checks for concepts with determined concept forms.
 * Composite checking part.
 *
 * @author Dennis Wagelaar
 * @version 0.1.18, 22 November 2003
 */

public class CompositeFormConstraintsVisitor extends ModelVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Parent implementation form constraint visitor.
   */
	private FormConstraintsVisitor parent;
  private ImplementationGenerator currentImplementation;
  private CompositeRole currentRole;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a CompositeFormConstraintsVisitor.
   * @param parent The parent visitor.
   */
  public CompositeFormConstraintsVisitor(FormConstraintsVisitor parent)
  {
    this.parent = parent;
  } //CompositeFormConstraintsVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Checks the concept form constraints for the composite.<BR>
   * PRE: The model must be defined for this composite.
   *
   * @param composite The composite to check the forms for.
   * @throws CoComposeException
   *   if one of the model's constraints are not fulfilled or
   *   if the concept form constraints are not met.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
    Model model = composite.getModel();
    ImplementationGenerator impl;
    boolean implWorks = false;

    if (log != null)
    {
      log.log("Checking concept form constraints for " + composite.getName());
      log.incLogLevel();
    }
    try
    {
      /* Precondition check -- disabled
      // check model
      if (model == null)
      {
        if (log != null)
          log.log("Error: " + composite.getName() + " has no model set");
        throw new Exception(composite.getName() + " has no model set");
      }
      End precondition check */
      // check if we have implementation generators --
      Enumeration impls = composite.getImplementations(model.getTargetLanguage());
      while (impls.hasMoreElements())
      {
        impl = (ImplementationGenerator) impls.nextElement();
        // check implementation generator concept form constraints
        try
        {
          impl.accept(this);
          implWorks = true;
        }
        catch (CoComposeException e)
        { // try other implementation generators
        }
      }
      if (!implWorks)
      {
        if (log != null)
          log.log("No implementation generator works for " + composite.getName());
        throw new CoComposeException("No implementation generator works for " +
          composite.getName(), composite);
      }
      // -- end implementation generator check
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitComposite

  /**
   * Check concept form constraints for this implementation generator.<BR>
   * PRE: The composite and model must be defined for this implementation generator.
   *
   * @param implementation The implementation generator to check.
   * @throws CoComposeException
   *   if one of the model's constraints are not fulfilled or
   *   if the concept form constraints are not met.
   */
  public void visitImplementationGenerator(ImplementationGenerator implementation)
    throws CoComposeException
  {
    Composite composite = implementation.getComposite();
    CompositeRole role;

    if (log != null)
    {
      log.log("Checking concept form constraints for " +
        implementation.getFilename());
      log.incLogLevel();
    }
    try
    {
      /* Precondition check -- disabled
      // check composite
      if (composite == null)
      {
        if (log != null)
          log.log("Error: implementation generator has no composite set");
        throw new Exception("Implementation generator has no composite set");
      }
      // check model
      if (model == null)
      {
        if (log != null)
          log.log("Error: implementation generator has no model set");
        throw new Exception("Implementation generator has no model set");
      }
      End precondition check */
      currentImplementation = implementation;
      // -- check for each role and each concept filling that role:
      Enumeration roles = composite.getRoles();
      while (roles.hasMoreElements())
      {
        role = (CompositeRole) roles.nextElement();
        checkCompositeRole(role, implementation);
      } //while roles
      // -- end constraints check
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitImplementationGenerator

  /**
   * Checks a CompositeRole object.
   * @param role The composite role to check.
   * @param implementation The implementation generator to check.
   * @throws CoComposeException if visiting operation failed.
   */
  public void checkCompositeRole(CompositeRole role, 
    ImplementationGenerator implementation) throws CoComposeException
  {
    Model model = role.getModel();
    Node node;
    currentRole = role;
    // check if any concepts filling this role use this role part's form
    for (int child = role.firstChild(); child >= 0; child = role.nextChild())
    {
      node = model.getNodeFromIndex(child);
      if (node instanceof EDU.utwente.CoCompose.lang.Concept)
      { // only check Concepts
        checkConcept((Concept) node, role, implementation);
      }
    } //for child
  } //checkCompositeRole

  /**
   * Checks a Concept object.
   * @param concept The concept to check.
   * @param role The composite role to check.
   * @param implementation The implementation generator to check.
   * @throws CoComposeException if visiting operation failed.
   */
  public void checkConcept(Concept concept, CompositeRole role, 
    ImplementationGenerator implementation) throws CoComposeException
  {
    Hashtable forms = parent.getForms();
    ImplementationRolePart rolepart;
    Enumeration roleParts = implementation.getRoleParts(role.getName());
    // only check concepts with forms determined
    if (!forms.containsKey(concept))
      return;
    // check if this role part works for this concept
    while (roleParts.hasMoreElements())
    {
      rolepart = (ImplementationRolePart) roleParts.nextElement();
      try
      {
        if (forms.get(concept).equals(rolepart.getConceptForm()))
        { // this concept uses the concept form of this role part, so
          // check this rolepart
          rolepart.accept(this);
        }
        else
        {
          throw new CoComposeException("Role part for " + 
            rolepart.getRoleName() + " as " + rolepart.getConceptForm() + 
            " did not work for " + concept.getName() + " as " + 
            forms.get(concept), concept);
        }
        return; // works, so no check needed for other role parts
      }
      catch (CoComposeException e)
      { // this role part didn't work for the concept, try next 
      }
    } //while roleParts
    if (log != null)
      log.log("No role part for " + 
        currentRole.getName() + " works for " + concept.getName() + " as " + 
        forms.get(concept));
    throw new CoComposeException("No role part for " + 
      currentRole.getName() + " works for " + concept.getName() + " as " + 
      forms.get(concept), concept);
  } //checkConcept

  /**
   * Check concept form constraints for this implementation generator's role
   * part.<BR>
   * PRE: The implementation generator, composite and model must be defined for this
   * role part.
   *
   * @param rolepart The implementation generator role part to check.
   * @throws CoComposeException
   *   if one of the model's constraints are not fulfilled or
   *   if the concept form constraints are not met.
   */
  public void visitImplementationRolePart(ImplementationRolePart rolepart)
    throws CoComposeException
  {
    Hashtable forms = parent.getForms();
    ImplementationGenerator implementation = rolepart.getImplementationGenerator();
    Composite composite = rolepart.getComposite();
    Model model = rolepart.getModel();
    CompositeRole role;
    String notEqualTo;
    String rolename;
    Node node;

    if (log != null)
    {
      log.log("Checking constraints of the role part for " +
        rolepart.getRoleName() + " as " + rolepart.getConceptForm());
      log.incLogLevel();
    }
    try
    {
      /* Precondition check -- disabled
      // check implementation generator
      if (implementation == null)
      {
        if (log != null)
          log.log("Error: role part for " + rolepart.getRoleName() +
            " has no implementation set");
        throw new Exception("Role part for " + rolepart.getRoleName() +
          " has no implementation set");
      }
      // check model
      if (model == null)
      {
        if (log != null)
          log.log("Error: role part for " + rolepart.getRoleName() +
            " has no model set");
        throw new Exception("Role part for " + rolepart.getRoleName() +
          " has no model set");
      }
      // check composite
      if (composite == null)
      {
        if (log != null)
          log.log("Error: role part for " + rolepart.getRoleName() +
            " has no composite set");
        throw new Exception("Role part for " + rolepart.getRoleName() +
          " has no composite set");
      }
      End precondition check */
      // check constraints for this rolepart --
      // find and check each role within composite which is used in constraints
      Enumeration roleNames = rolepart.getConstraintRoles();
      while (roleNames.hasMoreElements())
      {
        rolename = (String) roleNames.nextElement();
        role = composite.getRole(rolename);
        // check role
        if (role == null)
        {
          if (log != null)
            log.log("Role " + rolename + " not found in " + composite.getName());
          throw new CoComposeException("Role " + rolename + " not found in " +
            composite.getName(), composite);
        }
        Enumeration constraints = rolepart.getConstraints(rolename);
        while (constraints.hasMoreElements())
        {
          notEqualTo = (String) constraints.nextElement();
          if (log != null)
            log.log("Checking constraint " + rolename + " != " + notEqualTo);
          // check constraint for each Concept filling the role
          for (int child = role.firstChild(); child >= 0; child = role.nextChild())
          {
            node = model.getNodeFromIndex(child);
            if (node instanceof EDU.utwente.CoCompose.lang.Concept)
            { // only check Concepts
              if (forms.containsKey(node))
              { // only check concepts with forms determined
                if (forms.get((Concept) node).equals(notEqualTo))
                { // if concept form determined and equal to constraint, this
                  // role part cannot be used
                  if (log != null)
                    log.log("Constraints not fulfilled for the role part for " +
                      rolepart.getRoleName() + " as " + rolepart.getConceptForm());
                  throw new CoComposeException("Constraints not fulfilled for the role part for " +
                    rolepart.getRoleName() + " as " + rolepart.getConceptForm(), composite);
                }
              }
            } //if node
          } //for child
        } //while constraints
      } //while roleNames
      if (log != null)
        log.log("Constraints fulfilled for the role part for " +
          rolepart.getRoleName() + " as " + rolepart.getConceptForm());
      // -- end rolepart check
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitImplementationRolePart
} //class CompositeFormConstraintsVisitor
