/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Determines the concept forms of all concepts in a model.
 *
 * @author Dennis Wagelaar
 * @version 0.1.20, 4 April 2004
 */

public class DetermineFormsVisitor extends ModelVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // The mapping of concepts to sets of possible concept forms:
  // Hashtable<AbstractConcept |-> Set<String>>
  private Hashtable forms_ = new Hashtable();
  // The set of permutations of concept forms for the entire model:
  // Set<Hashtable<Concept |-> String>>
  private HashSet permutations_ = new HashSet();
  // The set of resulting models: Set<Model>
  private Set resultModels_ = null;
  /**
   * The model on which this visitor is applied.
   */
  protected Model thisModel = null;
  /**
   * The form checking visitor needed to check constraints
   */
  protected FormConstraintsVisitor constraintCheck = null;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a ModelVisitor.
   */
  public DetermineFormsVisitor()
  {
  } //DetermineFormsVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the set of concept forms for a concept.
   *
   * @param concept The concept to get the forms for.
   * @return The set of concept forms.
   */
  protected Set getConceptForms(AbstractConcept concept)
  {
    return (Set) forms_.get(concept);
  } //getConceptForms

  /**
   * Sets the set of concept forms for a concept.
   * @param concept - the concept to set the forms for.
   * @param forms - the Set<String> of concept forms.
   */
  protected void setConceptForms(AbstractConcept concept, Set forms)
  {
    forms_.put(concept, forms);
  } //setConceptForms

  /**
   * Sets the set of concept forms for a concept and generates permutations.
   * @param concept - the concept to set the forms for.
   * @param forms - the Set<String> of concept forms.
   * @throws CoComposeException - if no permutations left afterwards.
   */
  protected void setConceptForms(Concept concept, Set forms)
    throws CoComposeException
  {
    // set forms
    setConceptForms((AbstractConcept) concept, forms);
    // create permutations
    createPermutations(concept, forms);
  } //setConceptForms

  /**
   * Gets the concept form permutations for the entire model.
   * @return The concept form permutations: Set<Concept |-> String>.
   */
  protected Set getFormPermutations()
  {
    return permutations_;
  } //getFormPermutations

  /**
   * Sets the concept form permutations for the entire model.
   * @param permutations - the concept form permutations: Set<Concept |-> String>.
   */
  protected void setFormPermutations(Set permutations)
  {
    permutations_.clear();
    permutations_.addAll(permutations);
  } //setFormPermutations

  /**
   * Creates new concept forms permutations from Concept |-> form.
   * @param concept - the concept to create permutations for.
   * @param forms - the Set<String> of concept forms to create permutations for.
   * @throws CoComposeException - if no permutation is left afterwards.
   */
  protected void createPermutations(Concept concept, Set forms)
    throws CoComposeException
  {
    int permSize = getFormPermutations().size();
    if (log != null)
    {
      log.log("Creating permutations for " + concept.getName());
      log.incLogLevel();
      log.log(permSize + " permutations exist so far");
      log.log(concept.getName() + " has " + forms.size() + " possible forms");
    }
    try
    {
      // empty set of permutations, so make new set
      if (permSize == 0)
        createNewPermutations(concept, forms);
      // permutations already exist, so merge new concept forms
      else
        createAdditionalPermutations(concept, forms);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //createPermutations

  /**
   * Creates new concept forms permutations from Concept |-> form. Assumes no
   * concept form permutations have been determined so far.
   * @param concept - the concept to create permutations for.
   * @param forms - the Set<String> of concept forms to create permutations for.
   */
  protected void createNewPermutations(Concept concept, Set forms)
  {
    Set permutations = getFormPermutations();
    Iterator formsIt = forms.iterator();
    while (formsIt.hasNext())
    {
      String form = (String) formsIt.next();
      Hashtable newPermutation = new Hashtable();
      newPermutation.put(concept, form);
      permutations.add(newPermutation);
    }
  } //createNewPermutations

  /**
   * Creates new concept forms permutations from Concept |-> form. Assumes some
   * concept form permutations have been determined already.
   * @param concept - the concept to create permutations for.
   * @param forms - the Set<String> of concept forms to create permutations for.
   * @throws CoComposeException - if no permutation is left afterwards.
   */
  protected void createAdditionalPermutations(Concept concept, Set forms)
    throws CoComposeException
  {
    int formCount = forms.size();
    Set permutations = getFormPermutations();
    Iterator permsIt = getFormPermutations().iterator();
    int counter = 0, elimCounter = 0, removeCounter = 0;
    HashSet removePerms = new HashSet();

    if (formCount == 1)
    { // do simplified procedure for one single additional form
      while (permsIt.hasNext())
      {
        Hashtable permutation = (Hashtable) permsIt.next();
        String form = (String) forms.iterator().next();
        permutation.put(concept, form);
        counter++;
      } //while permsIt
    }
    else
    { // do full procedure for multiple extra forms
      HashSet newPermutations = new HashSet();
      while (permsIt.hasNext())
      {
        Hashtable permutation = (Hashtable) permsIt.next();
        int currentForm = 0;
        Iterator formsIt = forms.iterator();
        while (formsIt.hasNext())
        {
          currentForm++;
          String form = (String) formsIt.next();
          Hashtable newPermutation = (Hashtable) permutation.clone();
          newPermutation.put(concept, form);
          // check constraints for new permutation (incremental check)
          try
          { // check concept form constraints
            constraintCheck.setForms(newPermutation);
            constraintCheck.setOnlyNewConcept(concept);
            thisModel.accept(constraintCheck);
            newPermutations.add(newPermutation);
            counter++;
          }
          catch (CoComposeException e)
          { // constraints violated: don't add permutation
            elimCounter++;
            // if this was the last permutation and we still don't have a
            // feasible one, throw the exception
            if (!permsIt.hasNext() && !formsIt.hasNext() && counter == 0)
            {
              if (log != null)
                log.log("Error: " + e.getMessage());
              throw e;
            }
          }
        } //while formsIt
      } //while permsIt
      setFormPermutations(newPermutations);
      if (log != null)
      {
        log.log(elimCounter + " permutations were eliminated in advance");
        log.log("Model now has " + permutations.size() + " permutations");
      }
    }
  } //createAdditionalPermutations

  /**
   * Eliminates permutations that violate concept form constraints.
   * @throws CoComposeException if all permutations got eliminated.
   */
  protected void eliminatePermutations() throws CoComposeException
  {
    Set permutations, marked = new HashSet();
    Hashtable permutation;
    int counter = 0;

    if (log != null)
    {
      log.log("Eliminating invalid permutations");
      log.incLogLevel();
    }
    try
    {
      constraintCheck.setOnlyNewConcept(null);
      // eliminate invalid concept form permutations
      permutations = getFormPermutations();
      Iterator perms = permutations.iterator();
      while (perms.hasNext())
      {
        permutation = (Hashtable) perms.next();
        try
        { // check concept form constraints
          constraintCheck.setForms(permutation);
          thisModel.accept(constraintCheck);
        }
        catch (CoComposeException e)
        { // constraints violated: eliminate permutation
          if (permutations.size() == 1)
          { // if this was the last permutation, throw the error
            if (log != null)
              log.log("Error: " + e.getMessage());
            throw e;
          }
          else
          {
            marked.add(permutation);
            counter++;
          }
        }
      } //while
      // remove marked permutations - via strange detour...
      HashSet newPermutations = new HashSet(permutations);
      newPermutations.removeAll(marked);
      setFormPermutations(newPermutations);
      if (log != null)
      {
        log.log("Eliminated " + counter + " permutations");
        log.log("Model now has " + permutations.size() + " permutations left");
      }
    }
    finally
    {
      // force garbage collection
      System.gc();
      if (log != null)
        log.decLogLevel();
    }
  } //eliminatePermutations

  /**
   * Applies the concept form heuristics defined in the model to eliminate
   * permutations.
   */
  protected void applyHeuristics()
  {
    Set permutations, marked;
    Hashtable permutation, otherPerm, delete;
    FormConstraintsVisitor constraintCheck;
    Iterator perms, otherPerms;

    if (log != null)
    {
      log.log("Applying concept form heuristics to eliminate permutations");
      log.incLogLevel();
    }
    try
    {
      // compare each pair of permutations that only has one different concept
      // form.
      permutations = getFormPermutations();
      marked = new HashSet(); // set of permutations marked for deletion
      perms = permutations.iterator();
      while (perms.hasNext())
      {
        permutation = (Hashtable) perms.next();
        otherPerms = permutations.iterator();
        while (otherPerms.hasNext())
        {
          otherPerm = (Hashtable) otherPerms.next();
          // don't compare permutation to self
          if (permutation == otherPerm)
            continue;
          delete = comparePermutations(permutation, otherPerm);
          if (delete != null)
          {
            marked.add(delete);
          }
        } //while otherPerms
      } //while perms
      // remove marked permutations - via strange detour...
      HashSet newPermutations = new HashSet(permutations);
      newPermutations.removeAll(marked);
      setFormPermutations(newPermutations);
      if (log != null)
      {
        log.log("Eliminated " + marked.size() + " permutations");
        log.log("Model now has " + permutations.size() + " permutations left");
      }
    }
    finally
    {
      // force garbage collection
      System.gc();
      if (log != null)
        log.decLogLevel();
    }
  } //applyHeuristics

  /**
   * Compares two permutations and returns the one that should be deleted
   * according to the concept form heuristics. This only applies if the
   * permutations differ in only one concept form, otherwise the method returns
   * null.
   * @param perm1 - the first permutation Hashtable<Concept |-> String> to compare.
   * @param perm2 - the second permutation Hashtable<Concept |-> String> to compare.
   * @return the permutation Hashtable<Concept |-> String> to delete (first or second), or null.
   */
  protected Hashtable comparePermutations(Hashtable perm1, Hashtable perm2)
  {
    Concept concept;
    String form1, form2, form;
    int diffcount = 0;
    Enumeration concepts, formPrior;
    Iterator formsIt;
    Hashtable result = null;

    // check whether this visitor has been applied
    if (thisModel == null)
    {
      if (log != null)
        log.log("This DetermineFormsVisitor has not been applied to any model");
      return null;
    }
    // start compare
    // first check if the permutations differ only in one concept form
    concepts = perm1.keys();
    while ((concepts.hasMoreElements()) && (diffcount < 2))
    {
      concept = (Concept) concepts.nextElement();
      form1 = (String) perm1.get(concept);
      form2 = (String) perm2.get(concept);
      if (!form1.equals(form2))
      { // if forms differ, increase diffcount and apply heuristics
        diffcount++;
        formPrior = thisModel.getTransPriorities(thisModel.getTargetLanguage());
        while ((formPrior.hasMoreElements()) && (result == null))
        {
          form = (String) formPrior.nextElement();
          if (form1.equals(form))
          { // form1 is prioritized form
            result = perm2; // mark perm2 for deletion
          }
          else if (form2.equals(form))
          { // form2 is prioritized form
            result = perm1; // mark perm1 for deletion
          }
        } //while formPrior
      }
    } //while concepts
    // check if more than one difference found
    if (diffcount > 1)
    {
      return null;
    }
    return result;
  } //comparePermutations

  /**
   * Creates the resulting models using the valid permutations of concept
   * forms.
   */
  public void createResultingModels()
  {
    Set permutations = getFormPermutations();
    Hashtable permutation;
    Concept concept, conceptCopy;
    Model modelCopy;
    Iterator concepts;
    AbstractConcept tmpconcept;

    resultModels_ = new HashSet();
    Iterator perms = permutations.iterator();
    while (perms.hasNext())
    {
      permutation = (Hashtable) perms.next();
      modelCopy = (Model) thisModel.clone();
      concepts = thisModel.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      {
        tmpconcept = (AbstractConcept) concepts.next();
        if (tmpconcept instanceof Concept)
        {
          concept = (Concept) tmpconcept;
          // check whether permutation complete, if not return result so far
          if (!permutation.containsKey(concept))
            return;
          // retrieve copy concept
          conceptCopy = (Concept) modelCopy.getNodeFromId(concept.getId());
          conceptCopy.setForm((String) permutation.get(concept));
        }
      }
      resultModels_.add(modelCopy);
    }
  } //createResultingModels

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Retrieves the resulting models using the valid permutations of concept
   * forms.
   * @return the resulting set of models: Set<Model>.
   */
  public Collection getResultingModels()
  {
    // create (incomplete) result if not yet available
    if (resultModels_ == null)
      createResultingModels();
    return resultModels_;
  } //getResultingModels

  /**
   * Determines the concept forms for each concept in the model.
   * @param model - the model to visit.
   * @throws CoComposeException - if one of the model's constraints are not fulfilled or
   *         if not all concept forms could be determined.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    AbstractConcept concept;
    ModelVisitor postproc;
    String postprocClassname;
    Iterator models, concepts;

    if (log != null)
    {
      log.log("-- Started determine concept forms for " + model.getName());
      log.incLogLevel();
    }
    try
    {
      thisModel = model;
      // enforce model to be correct and complete
      model.checkCorrectness();
      model.checkCompleteness();
      // create constraint checker
      constraintCheck = new FormConstraintsVisitor();
      // determine forms for each concept
      concepts = model.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      {
        concept = (AbstractConcept) concepts.next();
        if (concept instanceof Concept)
        {
          concept.accept(this);
        }
      }
      // eliminate invalid concept forms using constraints
      eliminatePermutations();
      // apply heuristics to eliminate form permutations
      applyHeuristics();
      // create resulting models
      createResultingModels();
      // apply implementation language post-processing, if possible
      try
      {
        postprocClassname = "translate." + model.getTargetLanguage() +
          "FormsPostproc";
        postproc = (ModelVisitor) Class.forName(postprocClassname).newInstance();
        postproc.setLogger(log);
        models = getResultingModels().iterator();
        while (models.hasNext())
        {
          ((Model) models.next()).accept(postproc);
        }
      }
      catch (Exception e)
      {
        if (log != null)
          log.log("Target language specific postprocessing not (fully) applied (" +
            e.getMessage() + ")");
      }
    }
    finally
    {
      // force garbage collection
      System.gc();
      if (log != null)
      {
        log.decLogLevel();
        log.log("-- Completed determine concept forms for " + model.getName());
      }
    }
  } //visitModel

  /**
   * Determines the possible concept forms for this concept using the target
   * language of the model.<BR>
   * PRE: The model must be defined for this concept.
   * @param concept - the concept to determine the forms for.
   * @throws CoComposeException - if the concept forms could not be determined.
   */
  public void visitConcept(Concept concept) throws CoComposeException
  {
    Model model = concept.getModel();
    Set result = null;
    Set subresult = null;
    int child;
    Node node;
    CompositeRole role;
    Implementation impl;

    if (log != null)
    {
      log.log("Determining all possible concept forms for " + concept.getName());
      log.incLogLevel();
    }
    try
    {
      /* Precondition check -- disabled
      // check model
      if (model == null)
      {
        if (log != null)
          log.log("Error: " + concept.getName() + " has no model set");
        throw new CoComposeException(concept.getName() + " has no model set");
      }
      End precondition check */
      // check each role we fill --
      for (child = concept.firstChild(); child != -1; child = concept.nextChild())
      {
        node = model.getNodeFromIndex(child);
        if (node instanceof EDU.utwente.CoCompose.lang.CompositeRole)
        {
          role = (CompositeRole) node;
          if (log != null)
            log.log(concept.getName() + " is linked to " + role.getName());
          // retrieve concept forms from composite role
          role.accept(this);
          subresult = getConceptForms(role);
          if (result != null)
            // intersect with existing result
            result.retainAll(subresult);
          else
            // create new result
            result = new HashSet(subresult);
        }
      } //for child
      // -- end role check
      // check if we have implementations --
      if (concept.getImplementationCount() > 0)
      {
        subresult = new HashSet();
        if (log != null)
          log.log(concept.getName() + " has implementations");
        // process implementations
        Enumeration impls = concept.getImplementations(model.getTargetLanguage());
        while (impls.hasMoreElements())
        {
          impl = (Implementation) impls.nextElement();
          // add implementation concept form
          subresult.add(impl.getConceptForm());
          if (log != null)
            log.log("Adding concept form " + impl.getConceptForm());
        }
      }
      // merge with inherited result
      if (subresult != null)
      { // there are implementations
        if (result != null)
        { // there is an inherited result, so intersect
          result.retainAll(subresult);
        }
        else
        { // no inherited result, so set to subresult
          result = subresult;
        }
      }
      // -- end implementation check
      // check if we found concept forms
      if (result == null)
      {
        if (log != null)
          log.log("No concept form constraints found yet for " + concept.getName() + 
            " - adding all forms found in heuristics");
        result = new HashSet();
        Enumeration prior = model.getTransPriorities(model.getTargetLanguage());
        while (prior.hasMoreElements())
        {
          String form = (String) prior.nextElement();
          result.add(form);
        }
      }
      else if (result.isEmpty())
      {
        if (log != null)
          log.log("No concept forms found for " + concept.getName());
        throw new CoComposeException("No concept forms found for " +
          concept.getName(), concept);
      }
      setConceptForms(concept, result);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitConcept

  /**
   * Gets the possible concept forms for which the composite will translate
   * into the model's target language.<BR>
   * PRE: The model and composite must be defined for this composite role.
   * @param role - the composite role to get the forms for.
   * @throws CoComposeException - if the concept forms could not be retrieved.
   */
  public void visitCompositeRole(CompositeRole role) throws CoComposeException
  {
    Set result = new HashSet();
    Set subresult;
    Composite composite = role.getComposite();
    Model model = role.getModel();
    ImplementationGenerator impl;
    boolean implPresent = false;

    if (log != null)
    {
      log.log("Retrieving possible concept forms for the " + role.getName() + " role");
      log.incLogLevel();
    }
    try
    {
      /* Precondition check -- disabled
      // check composite
      if (composite == null)
      {
        if (log != null)
          log.log("Error: " + role.getName() + " role is not part of a composite");
        throw new CoComposeException(role.getName() + " role is not part of a composite");
      }
      // check model
      if (model == null)
      {
        if (log != null)
          log.log("Error: " + role.getName() + " has no model set");
        throw new CoComposeException(role.getName() + " has no model set");
      }
      End precondition check */
      // retrieve concept forms from all implementation generators
      Enumeration impls = composite.getImplementations(model.getTargetLanguage());
      while (impls.hasMoreElements())
      {
        implPresent = true;
        impl = (ImplementationGenerator) impls.nextElement();
        if (log != null)
          log.log("Adding concept forms of implementation generator " + impl.getFilename());
        subresult = impl.getRoleForms(role.getName(), log);
        result.addAll(subresult);
      }
      if (!implPresent)
      { // if no implementation generators present, no forms can be determined
        if (log != null)
          log.log("Error: no implementation generators defined for " +
            composite.getName() + ": try flattening first");
        throw new CoComposeException("No implementation generators defined for " +
          composite.getName() + ": try flattening first", composite);
      }
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
    setConceptForms(role, result);
  } //visitCompositeRole
} //class DetermineFormsVisitor
