/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002, 2003
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.util.*;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Flattens the model using available solution patterns.
 *
 * @author Dennis Wagelaar
 * @version 0.1.16, 31 October 2003
 * @deprecated
 */

public class FlattenVisitor extends ModelVisitor
{
  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  // If true, does full flatten, otherwise one level flatten.
  private boolean full_ = true;
  // flattened models
  private Vector flatModels_ = new Vector();
  // The set of composites that should not be flattened.
  private Set dontFlatten_ = new HashSet();
  // The set of composite types that should not be flattened.
  private Set dontFlattenTypes_ = new HashSet();
  // Determines flattening options.
  private FlatteningOptionsVisitor flatteningOptions = new FlatteningOptionsVisitor();
  // low-level composite cache (when checked, these can be skipped next time)
  // Format: composite types: Set<String>
  private Set lowLevelCompositesCache_ = new HashSet();
  /**
   * When false, flattening did not change model, so it was not needed.
   */
  protected boolean flattenNeeded = false;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a FlattenVisitor.
   */
  public FlattenVisitor()
  {
  } //FlattenVisitor

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets full flattening.
   *
   * @return If true, does full flattening, otherwise one level flattening.
   */
  protected boolean getFull()
  {
    return full_;
  } //getFull

  /**
   * Sets the flattened models so far.
   *
   * @param flatModels The collection of flattened models.
   */
  protected void setFlatModels(Collection flatModels)
  {
    flatModels_.clear();
    flatModels_.addAll(flatModels);
  } //setFlatModels

  /**
   * Adds a flattened model to the collection of flattened models.
   *
   * @param flatModel The flattened model to add.
   */
  protected void addFlatModel(Model flatModel)
  {
    flatModels_.add(flatModel);
  } //addFlatModel

  /**
   * Checks whether a composite should be flattened based on the set of
   * composites that need not be flattened.
   * @param composite The composite
   * @return True if and only if the composite should be flattened.
   */
  protected boolean dontFlatten(Composite composite)
  {
    return (dontFlatten_.contains(composite) ||
            dontFlattenTypes_.contains(composite.getType()));
  } //dontFlatten

  /**
   * Gets the set of composites that should not be flattened.
   *
   * @return The set of composites.
   */
  protected Set getDontFlatten()
  {
    return dontFlatten_;
  } //getDontFlatten

  /**
   * Adds a composite to the set of composites that should not be flattened.
   * This option is used to prevent infinite loops in flattening.
   *
   * @param composite The composite that should not be flattened.
   */
  protected void addDontFlatten(Composite composite)
  {
    dontFlatten_.add(composite);
  } //addDontFlatten

  /**
   * Sets the set of composites that should not be flattened.
   * This option is used to prevent infinite loops in flattening.
   *
   * @param composites The set of composites that should not be flattened.
   */
  protected void setDontFlatten(Set composites)
  {
    dontFlatten_.clear();
    dontFlatten_.addAll(composites);
  } //setDontFlatten

  /**
   * Gets the set of composite types that should not be flattened.
   *
   * @return The set of composite types.
   */
  protected Set getDontFlattenTypes()
  {
    return dontFlattenTypes_;
  } //getDontFlattenTypes

  /**
   * Adds a composite type to the set of composite types that should not be
   * flattened.
   *
   * @param type The composite type that should not be flattened.
   */
  protected void addDontFlattenType(String type)
  {
    dontFlattenTypes_.add(type);
  } //addDontFlattenType

  /**
   * Sets the set of composite types that should not be flattened.
   *
   * @param types The set of composite types that should not be flattened.
   */
  protected void setDontFlattenTypes(Set types)
  {
    dontFlattenTypes_.clear();
    dontFlattenTypes_.addAll(types);
  } //setDontFlattenTypes

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Sets the logger object.
   *
   * @param logger The object to log to
   */
  public void setLogger(Logger logger)
  {
    log = logger;
    flatteningOptions.setLogger(logger);
  } //setLogger

  /**
   * Sets full flattening.
   *
   * @param full If true, does full flattening, otherwise one level flattening.
   */
  public void setFull(boolean full)
  {
    full_ = full;
  } //setFull

  /**
   * Gets the flattened models so far.
   *
   * @return The collection of flattened models.
   */
  public Collection getFlatModels()
  {
    return flatModels_;
  } //getFlatModels

  /**
   * Flattens the model.
   *
   * @param model The model to flatten.
   * @throws CoComposeException
   *   if one of the model's constraints are not fulfilled or
   *   if the model could not be flattened.
   */
  public void visitModel(Model model) throws CoComposeException
  {
    Composite composite;
    Iterator models;
    Model flatModel;
    FlattenVisitor flatten;
    OrderCompositesVisitor order;
    Iterator composites;
    boolean full = getFull();
    Vector flatterModels = new Vector();

    if (log != null)
    {
      log.log("-- Started flattening for " + model.getName());
      log.incLogLevel();
    }
    try
    {
      // enforce model to be correct and complete
      model.checkCorrectness();
      model.checkCompleteness();
      // flatten to new model (copy of this)
      addFlatModel((Model) model.clone());
      // determine correct order for composites re. dependencies
      order = new OrderCompositesVisitor();
      order.setLogger(log);
      model.accept(order);
      composites = order.getComposites().iterator();
      // flatten each composite
      while (composites.hasNext())
      {
        composite = (Composite) composites.next();
        composite.accept(this);
      }
      if (!flattenNeeded && (log != null))
        log.log("Flattening not needed for " + model.getName());
      // if full flatten requested, flatten all resulting models
      if (full && flattenNeeded)
      {
        models = getFlatModels().iterator();
        while (models.hasNext())
        {
          flatModel = (Model) models.next();
          // create new flatten visitor for each level
          flatten = new FlattenVisitor();
          flatten.setLogger(log);
          flatten.setFull(full);
          flatten.setDontFlatten(getDontFlatten());
          flatten.setDontFlattenTypes(getDontFlattenTypes());
          flatten.lowLevelCompositesCache_ = lowLevelCompositesCache_;
          flatModel.accept(flatten);
          // save resulting models
          flatterModels.addAll(flatten.getFlatModels());
        }
        // set flatter models
        setFlatModels(flatterModels);
      }
    }
    finally
    {
      if (log != null)
      {
        log.decLogLevel();
        log.log("-- Completed flattening for " + model.getName());
      }
    }
  } //visitModel

  /**
   * Flattens a Composite object.
   * @param composite - the composite to flatten.
   * @throws CoComposeException - if the composite cannot be flattened for
   *         the target language of the model.
   */
  public void visitComposite(Composite composite) throws CoComposeException
  {
    Model model;
    Model origModel;
    Model modelCopy;
    SolutionPattern solution;
    Collection flatOptions;
    Iterator options;
    Vector newModels = new Vector();
    int optionCount;
    boolean nullSolution;

    if (dontFlatten(composite))
      return;
    if (log != null)
    {
      log.log("Flattening " + composite.getName());
      log.incLogLevel();
    }
    try
    {
      // check for target models
      Iterator targets = getFlatModels().iterator();
      if (!targets.hasNext())
      {
        if (log != null)
          log.log("No target models given for flattening " + composite.getName());
        throw new CoComposeException("No target models given for flattening " +
          composite.getName(), composite);
      }
      // flatten composite for each target model
      while (targets.hasNext())
      {
        model = (Model) targets.next();
        Composite targetComposite = (Composite) model.getNodeFromId(composite.getId());
        if (targetComposite.isDependent())
        {
          if (log != null)
            log.log(composite.getName() + " depends on published concepts and cannot be flattened yet");
        }
        else
        {
          // save original model state
          origModel = (Model) model.clone();
          // apply flattening options visitor to composite
          flatteningOptions.reset();
          flatteningOptions.setAppliedPatterns(model.getAppliedPatterns());
          flatteningOptions.setTargetLanguage(model.getTargetLanguage());
          composite.accept(flatteningOptions);
          // retrieve flattening options (always returns at least one option)
          flatOptions = flatteningOptions.getOptions();
          if (log != null)
            log.log(flatOptions.size() + " flattening options found for " +
              composite.getName());
          optionCount = 0;
          nullSolution = false;
          options = flatOptions.iterator();
          while (options.hasNext())
          {
            if (log != null)
              log.log("Applying option " + optionCount);
            if (optionCount != 0)
            { // create new target model for each option beyond the first
              model = (Model) origModel.clone();
              // add the new target model
              newModels.add(model);
            }
            // retrieve next option and apply it to the current target model
            solution = (SolutionPattern) options.next();
            if (solution != null)
            { // apply solution and remove this composite
              flattenNeeded = true;
              solution.apply(model, log);
            }
            else
            { // tag this composite, so it won't be flattened further (infinite loop)
              addDontFlatten((Composite) model.getAbstractConcept(
                composite.getName(), composite.getUniqueRealm().getName()));
              nullSolution = true;
            }
            // increase option count
            optionCount++;
          } //while options
          if (nullSolution && (optionCount < 1))
          { // this composite type has implementation generators only
            addDontFlattenType(composite.getType());
          }
        }
      } //while targets
      // add new models to target models list
      getFlatModels().addAll(newModels);
    }
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //visitComposite
} //class FlattenVisitor
