/**
 * Title:        CoCompose
 * Copyright:    Copyright (c) 2002 - 2004
 * Author:       Dennis Wagelaar
 * Organisation: University of Twente, Vrije Universiteit Brussel
 * Description:  CoCompose Design Language Tool
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (See the file "COPYING" that is included with this source distribution.)
 */

package EDU.utwente.CoCompose.translate;

import java.io.*;
import java.util.*;
import java.beans.PropertyChangeEvent;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.AttributesImpl;
import EDU.auburn.VGJ.graph.Node;
import EDU.utwente.CoCompose.lang.*;
import EDU.utwente.CoCompose.util.*;

/**
 * Represents a solution pattern for a Composite. It links to a file containing
 * the actual solution pattern; i.e. a CoCompose model XML file.
 *
 * @author Dennis Wagelaar
 * @version 0.1.20, 4 April 2004
 */

public class SolutionPattern implements Cloneable
{
  //////////////////////////////////////////////////////////////////////////////
  // Constants
  //////////////////////////////////////////////////////////////////////////////

  private static final String XMLNAME_SOLUTION = "SolutionPattern";

  //////////////////////////////////////////////////////////////////////////////
  // Attributes
  //////////////////////////////////////////////////////////////////////////////

  private String filename_ = new String();
  private DefaultUniqueRealm typeRealm_;
  private DefaultUniqueRealm instanceRealm_;
  private Composite composite_;
  private Model solutionModel_;
  /**
   * Cached Logger object.
   */
  protected Logger log;

  //////////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a SolutionPattern.
   */
  public SolutionPattern()
  {
    super();
  } //SolutionPattern

  //////////////////////////////////////////////////////////////////////////////
  // Private methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Initialises SolutionPattern for apply().
   * @param composite - the composite to replace by this solution pattern
   */
  private void initApply(Composite composite)
  {
    instanceRealm_ = new DefaultUniqueRealm();
    if ((composite.getUniqueRealm() != composite.getModel()) &&
        (composite.getUniqueRealm() != null))
    { // already part of a different namespace, so augment
      instanceRealm_.setName(composite.getUniqueRealm().getName() + '_' +
        composite.getName());
    }
    else
    { // create new namespace for this composite instance
      instanceRealm_.setName(composite.getName());
    }
  } //initApply

  /**
   * Retrieves the solution pattern model.
   * @param model - the model that is transformed
   * @return true if the model needed to be loaded from file
   * @throws CoComposeException - if the model could not be loaded
   */
  private boolean retrieveSolutionModel(Model model) throws CoComposeException
  {
    AppliedPatterns applied = model.getAppliedPatterns();
    int typecount = 0;

    if (!applied.containsPattern(getFilename()))
    { // this pattern has not been applied before, so a default part is needed
      // make new namespace for the composite type
      solutionModel_ = loadSolutionModel();
      typeRealm_ = new DefaultUniqueRealm();
      typeRealm_.setName(solutionModel_.getName());
      // check if type namespace exists
      while (getNamespaces(model).contains(typeRealm_.getName()))
      {
        typecount++;
        typeRealm_.setName(solutionModel_.getName() + typecount);
      }
      applied.addAppliedPattern(this, solutionModel_, typeRealm_);
      return true;
    }
    else
    { // retrieve used type name from cache
      solutionModel_ = (Model) applied.getSolutionModel(getFilename());
      typeRealm_ = (DefaultUniqueRealm) applied.getTypeRealm(getFilename());
      return false;
    }
  } //retrieveSolutionModel

  /**
   * Applies the solution pattern role parts.
   * @param model - the model to solve to.
   * @param composite - the composite being flattened.
   * @throws CoComposeException - if the role part could not be applied.
   */
  private void applyRPs(Model model, Composite composite) throws CoComposeException
  { // find and apply role part for each role filling
    Enumeration roles = composite.getRoles();
    while (roles.hasMoreElements())
    {
      CompositeRole role = (CompositeRole) roles.nextElement();
      for (int child = role.firstChild(); child >= 0; child = role.nextChild())
      {
        Node link = model.getNodeFromIndex(child);
        if (link instanceof EDU.utwente.CoCompose.lang.Concept)
        {
          applyRP(model, composite, role, (Concept) link);
        }
      }
    } //while
  } //applyRPs

  /**
   * Links the solution pattern role parts into the model.
   * @param model - the model to solve to.
   * @param composite - the composite being flattened.
   * @throws CoComposeException - if the role part could not be applied.
   */
  private void linkRPs(Model model, Composite composite) throws CoComposeException
  { // link role parts for each role filling
    Enumeration roles = composite.getRoles();
    while (roles.hasMoreElements())
    {
      CompositeRole role = (CompositeRole) roles.nextElement();
      for (int child = role.firstChild(); child >= 0; child = role.nextChild())
      {
        Node link = model.getNodeFromIndex(child);
        if (link instanceof EDU.utwente.CoCompose.lang.Concept)
        {
          linkRP(model, role, (Concept) link);
        }
      }
    } //while
  } //linkRPs

  /**
   * Links the solution pattern published concepts into the model.
   * @param model - the model to solve to.
   * @param composite - the composite being flattened.
   * @throws CoComposeException - if linking failed.
   */
  private void linkAllPublished(Model model, Composite composite)
    throws CoComposeException
  { // link published concepts for each link to the published concept
    Enumeration concepts = composite.getPublishedConcepts();
    while (concepts.hasMoreElements())
    {
      PublishedConcept published = (PublishedConcept) concepts.nextElement();
      for (int child = published.firstChild(); child >= 0; child = published.nextChild())
      {
        Node link = model.getNodeFromIndex(child);
        if (link instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
        {
          linkPublished(model, composite, published, (AbstractConcept) link);
        }
      }
      migratePublished(model, published, solutionModel_);
    } //while
  } //linkAllPublished

  /**
   * Adapts the name of the given concept using the role concept and the
   * concept that fills the role.
   * @param concept - the concept of which to adapt the name.
   * @param role - the solution pattern role of which the concept was part.
   * @throws CoComposeException if the new name cannot be set.
   */
  private void adaptName(AbstractConcept concept, SolutionRole role) 
    throws CoComposeException
  {
    String name, oldname, pubname = null;
    Enumeration concepts;
    AbstractConcept filler = role.getFilledBy();
    Composite composite = getComposite();

    if (concept instanceof EDU.utwente.CoCompose.lang.PublishedConcept)
      name = ((PublishedConcept) concept).getPublicName();
    else
      name = concept.getName();
    // check if name set, otherwise return
    if (name.length() == 0)
      return;
    oldname = name;
    // if a published concept refers to this concept, use its public name instead
    PublishedConcept pub = composite.getPublishedConcept(name);
    if ((pub != null) &&
        (composite.getRole(role.getName()).getChildren().numberOfElements() == 1))
    {
      pubname = pub.getPublicName();
      if (pubname.equals(""))
        pubname = null;
    }
    if (pubname != null)
    { // only use if the role is filled exactly once
      name = pub.getPublicName();
    }
    else if (name.startsWith(role.getName()))
    {
      name = filler.getName() + name.substring(role.getName().length());
    }
    else if (name.endsWith(role.getName()))
    {
      name = name.substring(0, name.length() - role.getName().length()) + filler.getName();
    }
    else
    {
      name += filler.getName();
    }
    if (concept instanceof EDU.utwente.CoCompose.lang.PublishedConcept)
      ((PublishedConcept) concept).setPublicName(name);
    else
      concept.setName(name);
    if (log != null)
      log.log("Adapted name " + oldname + " to " + name);
    // if Composite, then adapt published concept names as well (if public)
    if (concept instanceof EDU.utwente.CoCompose.lang.Composite)
    {
      concepts = ((Composite) concept).getPublishedConcepts();
      while (concepts.hasMoreElements())
      {
        adaptName((AbstractConcept) concepts.nextElement(), role);
      }
    }
  } //adaptName

  /**
   * Finds the concept in the solution pattern model that represents the given
   * published concept.
   * @param published - the published concept.
   * @param solutionModel - the solution pattern model.
   * @return the solution model concept that represents the given published
   *         concept.
   * @throws CoComposeException - if the solution model concept could not be found.
   */
  private AbstractConcept findSolutionConcept(PublishedConcept published,
    Model solutionModel) throws CoComposeException
  {
    AbstractConcept solutionConcept;
    // find concept in solution pattern
    solutionConcept = solutionModel.getAbstractConcept(published.getName(),
      solutionModel.getName());
    if (solutionConcept == null)
    {
      if (log != null)
        log.log(published.getName() + " concept not found in solution pattern of "
          + getComposite().getName());
      throw new CoComposeException(published.getName() +
        " concept not found in solution pattern of " +
        getComposite().getName(), getComposite());
    }
    return solutionConcept;
  } //findSolutionConcept

  /**
   * If the solution concept is published, we need to keep the copiedFrom history.
   * @param newConcept the newly created concept.
   * @param solConcept the solution pattern concept, which may be published.
   */
  private void updateCopiedFromHistory(AbstractConcept newConcept, 
    AbstractConcept solConcept)
  {
    PublishedConcept pub = getPublished(solConcept);
    if (pub != null)
    { // if the solution concept was published, we need to keep the copiedFrom history
      newConcept.getCopiedFrom().addAll(pub.getCopiedFrom());
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  // Protected methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Attempts to find the SolutionRole with the given name within the solution
   * pattern.
   * @param name - the SolutionRole name
   * @return the SolutionRole, if found
   * @throws CoComposeException if the SolutionRole is not found
   */
  protected SolutionRole findSolutionRole(String name) throws CoComposeException
  {
    if (solutionModel_ == null)
      throw new CoComposeException("Solution model not yet loaded when trying to find a SolutionRole");
    AbstractConcept tmpconcept =
      solutionModel_.getAbstractConcept(name, solutionModel_.getName());
    if (tmpconcept instanceof EDU.utwente.CoCompose.lang.SolutionRole)
    {
      return (SolutionRole) tmpconcept;
    }
    else
    {
      if (log != null)
        log.log(name + " role part not found in solution pattern of "
          + getComposite().getName());
      throw new CoComposeException(name +
        " role part not found in solution pattern of " +
        getComposite().getName(), getComposite());
    }
  } //findSolutionRole

  /**
   * Gets the namespaces present in the given model.
   * @param model - the model to look in.
   * @return the set of UniqueRealm name strings that make up each namespace.
   */
  protected Set getNamespaces(Model model)
  {
    HashSet namespaces = new HashSet();
    Node tmpnode;

    for (tmpnode = model.firstFreeNode(); tmpnode != null; tmpnode = model.nextFreeNode(tmpnode))
    {
      if (tmpnode instanceof EDU.utwente.CoCompose.lang.Concept)
      {
        namespaces.add(((Concept) tmpnode).getUniqueRealm().getName());
      }
      else if (tmpnode instanceof EDU.utwente.CoCompose.lang.Composite)
      {
        namespaces.add(((Composite) tmpnode).getUniqueRealm().getName());
      }
    }
    return namespaces;
  } //getNamespaces

  /**
   * Links a new AbstractConcept according to the links in the solution model.
   * @param model - the target model.
   * @param solModel - the source (solution pattern) model.
   * @param concept - the AbstractConcept to make links for.
   * @param solConcept - the source AbstractConcept to read the links from.
   * @param instantiatedBy - the concepts to link should be instantiated by
   *        the role part for this. Ignored if null.
   * @throws CoComposeException - if one of the links could not be made.
   */
  protected void linkAbstractConcept(Model model, Model solModel,
    AbstractConcept concept, AbstractConcept solConcept,
    AbstractConcept instantiatedBy) throws CoComposeException
  {
    Node solnode;
    AbstractConcept link;
    Composite clink;
    Vector links = new Vector();
    Collection concepts, composites;
    Iterator conceptsIt, compositesIt;
    SolutionRole partOf;
    String namespace = null;

    if (log != null)
    {
      log.log("Linking " + concept.getName() + " into the model");
      log.incLogLevel();
    }
    try
    {
      for (int node = solConcept.firstChild(); node >= 0; node = solConcept.nextChild())
      {
        solnode = solModel.getNodeFromIndex(node);
        if (solnode instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
        {
          // find PartOf
          partOf = ((AbstractConcept) solnode).getPartOf();
          // determine if we have to link to a default part or role part
          if (partOf == null)
          { // link to default part
            // namespace is type model name
            namespace = typeRealm_.getName();
            concepts = new HashSet();
            if (solnode instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
            { // find composite in namespace first
              link = model.getAbstractConcept(
                ((CompositePartConcept) solnode).getComposite().getName(),
                namespace);
              if (link instanceof EDU.utwente.CoCompose.lang.Composite)
              { // found composite, now retrieve composite part concept
                namespace += "." + ((Composite) link).getName();
                link = ((Composite) link).getPartConcept(
                  ((AbstractConcept) solnode).getName());
              }
              else
              {
                link = null;
              }
            }
            else
            { // retrieve by namespace only
              link = model.getAbstractConcept(
                ((AbstractConcept) solnode).getName(), namespace);
            }
            if (link != null)
              concepts.add(link);
          }
          else
          { // link to role part
            // get link to role parts
            namespace = instanceRealm_.getName() + " (" + partOf.getName() + ")";
            if (solnode instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
            { // find composite in namespace first
              composites = model.getAbstractConcepts(
                ((CompositePartConcept) solnode).getComposite(),
                instanceRealm_);
              concepts = new HashSet();
              compositesIt = composites.iterator();
              while (compositesIt.hasNext())
              {
                clink = (Composite) compositesIt.next();
                namespace += "." + clink.getName();
                link = clink.getPartConcept(((AbstractConcept) solnode).getName());
                concepts.add(link);
              }
            }
            else
            { // retrieve by namespace only
              concepts = model.getAbstractConcepts((AbstractConcept) solnode,
                instanceRealm_);
            }
          }
          // add links if found
          if (concepts.size() > 0)
          {
            conceptsIt = concepts.iterator();
            while (conceptsIt.hasNext())
            {
              link = (AbstractConcept) conceptsIt.next();
              if ((instantiatedBy == null) ||
                  (link.getInstantiatedBy() == instantiatedBy) ||
                  (partOf != solConcept.getPartOf()))
              {
                links.add(link);
                if (log != null)
                  log.log("Linked " + concept.getName() + " to " + link.getName() +
                    " in namespace " + namespace);
              }
              else
              {
                if (log != null)
                  log.log("Link " + link.getName() +
                    " in namespace " + namespace + " not instantiated by " +
                    instantiatedBy.getName());
              }
            }
          }
          else
          {
            if (log != null)
              log.log("Could not find link yet for " + concept.getName() +
                " to " + ((AbstractConcept) solnode).getName() +
                " in namespace " + namespace + " - Delayed");
          }
        } //if solnode
      } // for node
      // link retrieved nodes
      model.linkNodes(concept, links);
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //linkAbstractConcept

  /**
   * Links a new Composite according to the links in the solution model. Links
   * the roles as well.
   * @param model - the target model.
   * @param solModel - the source (solution pattern) model.
   * @param composite - the Composite to make links for.
   * @param solComposite - the source Composite to read the links from.
   * @param instantiatedBy - the concepts to link should be instantiated by
   *        the role part for this. Ignored if null.
   * @throws CoComposeException - if one of the links could not be made.
   */
  protected void linkComposite(Model model, Model solModel,
    Composite composite, Composite solComposite, AbstractConcept instantiatedBy)
    throws CoComposeException
  {
    CompositePartConcept concept, solConcept;

    // link composite itself
    linkAbstractConcept(model, solModel, composite, solComposite, instantiatedBy);
    // link composite part concepts
    Enumeration concepts = composite.getPartConcepts();
    Enumeration solConcepts = solComposite.getPartConcepts();
    while (concepts.hasMoreElements() && solConcepts.hasMoreElements())
    {
      concept = (CompositePartConcept) concepts.nextElement();
      solConcept = (CompositePartConcept) solConcepts.nextElement();
      linkAbstractConcept(model, solModel, concept, solConcept, instantiatedBy);
    }
  } //linkComposite
  
  /**
   * Determines the FilledBy property for each SolutionRole, if possible. If
   * a SolutionRole is filled by more than one Concept, then the FilledBy
   * property will not be set.
   * @param model - the model to solve to.
   * @param composite - the composite being flattened.
   * @param solutionModel - the model containing the solution pattern.
   * @throws CoComposeException - if a solution role could not be found.
   */
  protected void determineFilledBy(Model model, Composite composite,
    Model solutionModel) throws CoComposeException
  {
    Enumeration roles;
    CompositeRole role;
    EDU.auburn.VGJ.graph.Set linkedNodes;
    AbstractConcept concept, tmpconcept;
    SolutionRole solutionRole = null;
    int index;

    if (log != null)
    {
      log.log("Determining FilledBy properties for solution roles");
      log.incLogLevel();
    }
    try
    {
      // find each solution role and set FilledBy if possible
      roles = composite.getRoles();
      while (roles.hasMoreElements())
      {
        role = (CompositeRole) roles.nextElement();
        linkedNodes = role.getChildren();
        // if there is only one concept linked, we can set FilledBy safely
        if (linkedNodes.numberOfElements() == 1)
        {
          index = linkedNodes.first();
          concept = (AbstractConcept) model.getNodeFromIndex(index);
	        // find solution role in solution pattern
	        tmpconcept = solutionModel.getAbstractConcept(role.getName(), solutionModel.getName());
          if (tmpconcept instanceof EDU.utwente.CoCompose.lang.SolutionRole)
	          solutionRole = (SolutionRole) tmpconcept;
          if (solutionRole == null)
          {
            if (log != null)
              log.log(role.getName() + " role part not found in solution pattern of "
              + composite.getName());
            throw new CoComposeException(role.getName() +
              " role part not found in solution pattern of " +
              composite.getName(), composite);
          }
          // set FilledBy property
          if (log != null)
            log.log(solutionRole.getName() + " FilledBy property set to " +
            concept.getName());
          solutionRole.setFilledBy(concept);
        }
      } //while
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //determineFilledBy

  /**
   * Applies the solution pattern default part.
   * @param model - the model to solve to.
   * @param composite - the composite being flattened.
   * @throws CoComposeException - if the default part could not be applied.
   */
  protected void applyDP(Model model, Composite composite)
    throws CoComposeException
  {
    Concept solConcept, newConcept;
    Composite solComposite, newComposite;
    Set insertedConcepts = new HashSet();
    Iterator inserted;

    if (log != null)
    {
      log.log("Applying default part");
      log.incLogLevel();
    }
    try
    {
      // apply default part for concepts
      for (solConcept = solutionModel_.firstConcept(); solConcept != null; 
        solConcept = solutionModel_.nextConcept(solConcept))
      {
        // all concepts without their PartOf relation filled in belong
        // to the default part.
        if (solConcept.getPartOf() == null)
        {
          newConcept = new Concept();
          updateCopiedFromHistory(newConcept, solConcept);
          newConcept.setUniqueRealm(typeRealm_);
          newConcept.copyAttributes(solConcept);
          // insert and link new concept
          model.insertNode(newConcept);
          if (log != null)
            log.log("Inserting concept " + newConcept.getName());
          linkAbstractConcept(model, solutionModel_, newConcept, solConcept, null);
          insertedConcepts.add(newConcept);
        }
      } //for solConcept
      // apply default part for composites and links
      for (solComposite = solutionModel_.firstComposite(); solComposite != null;
        solComposite = solutionModel_.nextComposite(solComposite))
      {
        // all composites without their PartOf relation filled in belong
        // to the default part.
        if (solComposite.getPartOf() == null)
        {
          newComposite = new Composite();
          newComposite.setUniqueRealm(typeRealm_);
          newComposite.copyAttributes(solComposite);
          // insert and link new composite
          model.insertNode(newComposite);
          if (log != null)
            log.log("Inserting composite " + newComposite.getName());
          linkComposite(model, solutionModel_, newComposite, solComposite, null);
        }
      } //for solComposite
      // copy all implementation constraints now that all concepts are present
      inserted = insertedConcepts.iterator();
      while (inserted.hasNext())
      {
        newConcept = (Concept) inserted.next();
        newConcept.copyConstraints((Concept) newConcept.getLastCopiedFrom());
      } //while inserted
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //applyDP

  /**
   * Applies the solution pattern role part for the given concept.
   * @param model - the model to solve to.
   * @param composite - the composite being flattened.
   * @param role - the role for which the role part will be applied.
   * @param concept - the concept filling the role.
   * @throws CoComposeException - if the role part could not be applied.
   */
  protected void applyRP(Model model, Composite composite, CompositeRole role,
    Concept concept) throws CoComposeException
  {
    Concept solConcept, newConcept;
    Composite solComposite, newComposite;
    AbstractConcept oldFilledBy;
    SolutionRole solutionRole = null;
    Set insertedConcepts = new HashSet();
    Iterator inserted;

    if (log != null)
    {
      log.log("Applying " + role.getName() + " role part for " + concept.getName());
      log.incLogLevel();
    }
    try
    {
      solutionRole = findSolutionRole(role.getName());
      // set FilledBy property
      oldFilledBy = solutionRole.getFilledBy();
      solutionRole.setFilledBy(concept);
      // apply role part for concepts
      for (solConcept = solutionModel_.firstConcept(); solConcept != null;
        solConcept = solutionModel_.nextConcept(solConcept))
      {
        // all concepts with their PartOf relation filled in by solutionRole
        if (solConcept.getPartOf() == solutionRole)
        {
          newConcept = new Concept();
          updateCopiedFromHistory(newConcept, solConcept);
          newConcept.copyAttributes(solConcept);
          // adapt name to the concept filling the role part
          adaptName(newConcept, solutionRole);
          newConcept.setUniqueRealm(instanceRealm_);
          newConcept.setInstantiatedBy(concept);
          // insert and link new concept
          model.insertNode(newConcept);
          if (log != null)
            log.log("Inserting concept " + newConcept.getName());
          linkAbstractConcept(model, solutionModel_, newConcept, solConcept, concept);
          insertedConcepts.add(newConcept);
        }
      } //for solConcept
      // apply role part for composites and links
      for (solComposite = solutionModel_.firstComposite(); solComposite != null;
        solComposite = solutionModel_.nextComposite(solComposite))
      {
        // all composites with their PartOf relation filled in by solutionRole
        if (solComposite.getPartOf() == solutionRole)
        {
          newComposite = new Composite();
          newComposite.copyAttributes(solComposite);
          // adapt name to the concept filling the role part
          adaptName(newComposite, solutionRole);
          newComposite.setUniqueRealm(instanceRealm_);
          newComposite.setInstantiatedBy(concept);
          // insert and link new composite
          model.insertNode(newComposite);
          if (log != null)
            log.log("Inserting composite " + newComposite.getName());
          linkComposite(model, solutionModel_, newComposite, solComposite, concept);
        }
      } //for solComposite
      // copy all implementation constraints now that all concepts are present
      inserted = insertedConcepts.iterator();
      while (inserted.hasNext())
      {
        newConcept = (Concept) inserted.next();
        newConcept.copyConstraints((Concept) newConcept.getLastCopiedFrom());
      } //while inserted
      // unset FilledBy property
      solutionRole.setFilledBy(oldFilledBy);
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //applyRP

  /**
   * Links the solution pattern role part into the model for the given concept.
   * @param model - the model to solve to.
   * @param role - the role for which the role part will be applied.
   * @param concept - the concept filling the role.
   * @throws CoComposeException - if the role part could not be applied.
   */
  protected void linkRP(Model model, CompositeRole role, Concept concept)
    throws CoComposeException
  {
    AbstractConcept tmpconcept, oldFilledBy;
    SolutionRole solutionRole = null;
    Composite composite = (Composite) model.getNodeFromId(getComposite().getId());

    if (log != null)
    {
      log.log("Linking " + role.getName() + " role part for " + concept.getName());
      log.incLogLevel();
    }
    try
    {
      // find role concept in solution pattern
      tmpconcept = solutionModel_.getAbstractConcept(role.getName(), solutionModel_.getName());
      if (tmpconcept instanceof EDU.utwente.CoCompose.lang.SolutionRole)
        solutionRole = (SolutionRole) tmpconcept;
      if (solutionRole == null)
      {
        if (log != null)
          log.log(role.getName() + " role part not found in solution pattern of "
            + composite.getName());
        throw new CoComposeException(role.getName() +
          " role part not found in solution pattern of " +
          composite.getName(), composite);
      }
      // set FilledBy property
      oldFilledBy = solutionRole.getFilledBy();
      solutionRole.setFilledBy(concept);
      // link the role part to the role filling concept
      linkAbstractConcept(model, solutionModel_, concept, solutionRole, concept);
      // unset FilledBy property
      solutionRole.setFilledBy(oldFilledBy);
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //linkRP

  /**
   * Links the solution pattern published concept into the model for the given
   * concept.
   * @param model - the model to solve to.
   * @param composite - the composite being flattened.
   * @param published - the published concept to be replaced.
   * @param concept - the concept linked to the published concept.
   * @throws CoComposeException - if linking failed.
   */
  protected void linkPublished(Model model, Composite composite,
    PublishedConcept published, AbstractConcept concept) throws CoComposeException
  {
    AbstractConcept solutionConcept;
    Node target;
    AbstractConcept tconcept;
    Collection targets = new Vector();
    UniqueRealm targetRealm;

    if (log != null)
    {
      log.log("Linking " + published.getName() + " published concept to " +
        concept.getName());
      log.incLogLevel();
    }
    try
    {
      solutionConcept = findSolutionConcept(published, solutionModel_);
      // find all concepts copied from the solutionConcept and link those
      for (target = model.firstNode(); target != null;
        target = model.nextNode(target))
      {
        if (target instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
        {
          tconcept = (AbstractConcept) target;
          if (tconcept instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
          { // use composite's realm
            targetRealm = ((CompositePartConcept) tconcept).getComposite().getUniqueRealm();
          }
          else
          {
            targetRealm = tconcept.getUniqueRealm();
          }
          if (targetRealm.getName().equals(instanceRealm_.getName()))
          { // search for copy of the solution concept
            if (tconcept.isCopiedFrom(solutionConcept))
            { // this is a copy representing the published concept, so link
              targets.add(tconcept);
              if (log != null)
                log.log("Linked " + concept.getName() + " to " + tconcept.getName());
            }
          }
          else if (targetRealm.getName().equals(typeRealm_.getName()))
          { // search for concept with same name
            if (tconcept.getPublicName().equals(solutionConcept.getPublicName()))
            { // this is a copy representing the published concept, so link
              targets.add(tconcept);
              if (log != null)
                log.log("Linked " + concept.getName() + " to " + tconcept.getName());
            }
          }
        }
      } //for
      // remove this link and then ass new link(s)
      model.removeEdge(published.getIndex(), concept.getIndex());
      model.linkNodes(concept, targets);
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //linkPublished

  /**
   * Migrates the implementation constraints from the given published concept
   * to the implementing concept.
   * @param model - the model to solve to.
   * @param published - the published concept from which to migrate.
   * @param solutionModel - the model containing the solution pattern.
   * @throws CoComposeException - if migration failed.
   */
  protected void migratePublished(Model model, PublishedConcept published,
    Model solutionModel) throws CoComposeException
  {
    AbstractConcept solutionConcept;
    Node target;
    AbstractConcept tconcept;
    Composite composite = (Composite) model.getNodeFromId(getComposite().getId());
    Collection targets = new Vector();
    UniqueRealm targetRealm;

    if (log != null)
    {
      log.log("Migrating implementation constraints for the " +
        published.getName() + " published concept");
      log.incLogLevel();
    }
    try
    {
      solutionConcept = findSolutionConcept(published, solutionModel);
      // find all concepts copied from the solutionConcept and migrate to those
      for (target = model.firstNode(); target != null;
        target = model.nextNode(target))
      {
        if (target instanceof EDU.utwente.CoCompose.lang.AbstractConcept)
        {
          tconcept = (AbstractConcept) target;
          if (tconcept instanceof EDU.utwente.CoCompose.lang.CompositePartConcept)
          { // use composite's realm
            targetRealm = ((CompositePartConcept) tconcept).getComposite().getUniqueRealm();
          }
          else
          {
            targetRealm = tconcept.getUniqueRealm();
          }
          if (targetRealm.getName().equals(instanceRealm_.getName()))
          { // search for copy of the solution concept
            if (tconcept.isCopiedFrom(solutionConcept))
            { // this is a copy representing the published concept, so migrate
              migrateImplementationConstraints(model, published, tconcept);
            }
          }
          else if (targetRealm.getName().equals(typeRealm_.getName()))
          { // search for concept with same name
            if (tconcept.getPublicName().equals(solutionConcept.getPublicName()))
            { // this is a copy representing the published concept, so migrate
              migrateImplementationConstraints(model, published, tconcept);
            }
          }
        }
      } //for
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //migratePublished

  /**
   * Migrates concept implementation constraints from the source concept to
   * the target concept.
   * @param model - the model that is being transformed.
   * @param source - the source concept to migrate from.
   * @param target - the target concept.
   * @throws CoComposeException - if migration failed.
   */
  protected void migrateImplementationConstraints(Model model,
    AbstractConcept source, AbstractConcept target) throws CoComposeException
  {
    Concept concept;
    Enumeration impls;
    Implementation impl;
    Hashtable constraints;
    PropertyChangeEvent evt;
    String sourceName, targetName;
    Iterator concepts;
    AbstractConcept tmpconcept;

    if (log != null)
    {
      log.log("Migrating Implementation constraints from " + source.getName()
        + " to " + target.getName());
      log.incLogLevel();
    }
    try
    {
      evt = Implementation.CreateRenameEvent(source, target);
      concepts = model.getAbstractConcepts().iterator();
      while (concepts.hasNext())
      { // check for each concept in the model whether it listens to the
        // published concept:
        tmpconcept = (AbstractConcept) concepts.next();
        if (tmpconcept instanceof Concept)
        {
          concept = (Concept) tmpconcept;
          impls = concept.getImplementations();
          while (impls.hasMoreElements())
          {
            impl = (Implementation) impls.nextElement();
            constraints = impl.getConstraints();
            if (constraints.keySet().contains(source))
            { // check for applicable constraint
            	if (log != null)
                log.log("Migrating constraints for " + concept.getName());
              impl.addConstraint(target, (Collection) constraints.get(source));
              impl.propertyChange(evt);
              impl.removeConstraint(source);
            }
          } //while
        }
      } //for
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //migrateImplementationConstraints

  /**
   * Loads the solution pattern model and returns it.
   * @return the model.
   * @throws CoComposeException - if the model could not be loaded.
   */
  protected Model loadSolutionModel() throws CoComposeException
  {
    Model solutionModel = null;

    File pathF = new File(getFilename());
    String path = pathF.getAbsolutePath().replace('\\', '/');
    String name = pathF.getName();
    if (name.endsWith(".xml"))
      name = name.substring(0, name.length() - 4);
    if (log != null)
    {
      log.log("Loading solution pattern model " + path);
    }
    try
    {
      solutionModel = new Model(path, name);
    }
    catch (Exception e)
    {
      throw new CoComposeException("Could not load solution pattern " + path +
        ": " + e.getMessage());
    }
    return solutionModel;
  } //loadSolutionModel

  /**
   * Retrieves the PublishedConcept that represents the given concept, if any.
   * @param concept the concept that is represented by a published concept.
   * @return the PublishedConcept that represents the given concept, if any.
   */
  protected PublishedConcept getPublished(AbstractConcept concept)
  {
    return getComposite().getPublishedConcept(concept.getName());
  }

  //////////////////////////////////////////////////////////////////////////////
  // Public methods
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Gets the name of the file that contains the actual solution pattern.
   * @return the file name.
   */
  public String getFilename()
  {
    return filename_;
  } //getFilename

  /**
   * Sets the name of the file that contains the actual solution pattern.
   * @param filename - the file name.
   */
  public void setFilename(String filename)
  {
    // compensate for missing "./" at beginning (MacOS X)
    if (!filename.startsWith("."))
      filename = "./" + filename.replace('\\', '/');
    filename_ = filename;
  } //setFilename

  /**
   * Gets the composite to which this solution pattern belongs.
   * @return the composite that owns this solution pattern.
   */
  public Composite getComposite()
  {
    return composite_;
  } //getComposite

  /**
   * Sets the composite to which this solution pattern belongs.
   * @param composite - the composite that owns this solution pattern.
   */
  public void setComposite(Composite composite)
  {
    composite_ = composite;
  } //setComposite

  /**
   * Copies the attributes from the given solution pattern to self. It doesn't
   * copy the composite link.
   * @param fromSolution - the solution pattern to copy from.
   */
  public void copyAttributes(SolutionPattern fromSolution)
  {
    setFilename(fromSolution.getFilename());
  } //copyAttributes

  /**
   * Create XML from this object using a SAX DefaultHandler.
   * @param handler - the SAX DefaultHandler.
   * @throws SAXException - if an XML handler error occurs.
   */
  public void toXML(DefaultHandler handler) throws SAXException
  {
    // Create attributes
    AttributesImpl atts = new AttributesImpl();
    atts.addAttribute(null, null, "Filename", "CDATA", getFilename());

    // Write XML
    handler.startElement(null, null, XMLNAME_SOLUTION, atts);
    handler.endElement(null, null, XMLNAME_SOLUTION);
  } //toXML

  /**
   * Applies the solution pattern for the composite in the given model.
   * @param model - the model context for the composite.
   * @param logger - the object to log to.
   * @throws CoComposeException - if the solution pattern cannot be applied.
   */
  public void apply(Model model, Logger logger)
    throws CoComposeException
  {
    // Retrieve the composite from the model context
    Composite composite = (Composite) model.getNodeFromId(getComposite().getId());
    log = logger;

    if (log != null)
    {
      log.log("Applying solution pattern " + getFilename());
      log.incLogLevel();
    }
    try
    {
      initApply(composite);
      if (retrieveSolutionModel(model))
      { // find and apply default part if it exists
        applyDP(model, composite);
      }
      determineFilledBy(model, composite, solutionModel_);
      applyRPs(model, composite);
      linkRPs(model, composite);
      linkAllPublished(model, composite);
      if (log != null)
        log.log("Removing solved composite " + composite.getName() +
          " from model " + model.getName());
      model.removeNodeFromId(composite.getId());
    } //try
    finally
    {
      if (log != null)
        log.decLogLevel();
    }
  } //apply

  /**
   * Accepts a ModelVisitor.
   * @param visitor - the visitor object to accept.
   * @throws CoComposeException - if visiting operation failed.
   */
  public void accept(ModelVisitor visitor) throws CoComposeException
  {
    visitor.visitSolutionPattern(this);
  } //accept
} //class SolutionPattern
